"""
Serialize a model once it has been built
Before solving and/or after solving. Clone operations executed. Scales not performed.

It could be something like:

Metadata:
{
{,
Parameters:
[
 {
 },
],
CodeHierarchies:
[
 {
 },
],
Observers:
[
 {
 },
],
InterfaceTypes:
[
 {
 },
],
InterfaceTypeConverts: (if conversions not done already)
[
 {
 },
],
Processors:
[
 {
  Interfaces:
  [
   {...,
    Observations:
    [
     {
     },
    ]
   }
  ]
 }
],
Relationships:
[
 {
  Origin: {name, uuid}
  Destination: {name, uuid}
  Type
 }
]
"""
import json
from collections import OrderedDict

import jsonpickle
from typing import Dict, List, Union, Optional, Any

from nexinfosys.common.helper import create_dictionary, CustomEncoder, values_of_nested_dictionary
from nexinfosys.model_services import State
from nexinfosys.models.musiasem_concepts import Processor, Parameter, Hierarchy, Taxon, Observer, FactorType, \
    ProcessorsRelationPartOfObservation, ProcessorsRelationUpscaleObservation, ProcessorsRelationIsAObservation, \
    FactorsRelationDirectedFlowObservation, FactorTypesRelationUnidirectionalLinearTransformObservation
from nexinfosys.solving import BasicQuery


JsonStructureType = Dict[str, Optional[Union[type, "JsonStructureType"]]]


def objects_list_to_string(objects_list: List[object], object_type: type) -> str:
    str_list = []

    if objects_list:
        if object_type is Hierarchy:
            # Just get the Hierarchy objects of type Taxon
            objects_list = [o for o in objects_list if o.hierarchy_type == Taxon]
            # Sort the list to show the "code lists" first
            objects_list.sort(key=lambda h: not h.is_code_list)

        for obj in objects_list:
            str_list.append(json.dumps(obj, cls=CustomEncoder))

    return ", ".join(str_list)


def create_json_string_from_objects(objects: Dict[type, List[object]], json_structure: JsonStructureType) -> str:

    def json_structure_to_string(sections_and_types: JsonStructureType) -> str:
        str_list = []
        for section_name, output_type in sections_and_types.items():
            if not isinstance(output_type, dict):
                str_list.append(f'"{section_name}": [{objects_list_to_string(objects.get(output_type), output_type)}]')
            else:
                str_list.append(f'"{section_name}": {{{json_structure_to_string(output_type)}}}')

        return ", ".join(str_list)

    return json_structure_to_string(json_structure)


# ------ MAIN FUNCTION ------

def model_to_json(state: State, structure: Dict):
    # Get objects from state
    query = BasicQuery(state)
    objects = query.execute(values_of_nested_dictionary(structure), filt="")

    return create_json_string_from_objects(objects, structure)


def export_model_to_json(state: State) -> str:

    # Get Metadata dictionary
    metadata = state.get("_metadata")
    metadata_string = None
    if metadata:
        # Change lists of 1 element to a simple value
        metadata = {k: v[0] if len(v) == 1 else v for k, v in metadata.items()}
        metadata_string = '"Metadata": ' + json.dumps(metadata, cls=CustomEncoder)

    json_structure: JsonStructureType = OrderedDict(
        {"Parameters": Parameter,
         "CodeHierarchies": Hierarchy,
         "Observers": Observer,
         "InterfaceTypes": FactorType,
         "InterfaceTypeConverts": FactorTypesRelationUnidirectionalLinearTransformObservation,
         "Processors": Processor,
         "Relationships": OrderedDict(
             {"PartOf": ProcessorsRelationPartOfObservation,
              "Upscale": ProcessorsRelationUpscaleObservation,
              "IsA": ProcessorsRelationIsAObservation,
              "DirectedFlow": FactorsRelationDirectedFlowObservation
              }
         )
         }
    )

    json_string = model_to_json(state, json_structure)

    if metadata_string:
        json_string = f'{metadata_string}, {json_string}'

    return f'{{{json_string}}}'




import json
import logging
from collections import OrderedDict

import pandas as pd
import numpy as np

import nexinfosys
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.common.helper import obtain_dataset_metadata, strcmp, create_dictionary, \
    augment_dataframe_with_mapped_columns, translate_case
from nexinfosys.models.musiasem_concepts import Hierarchy
from nexinfosys.models.musiasem_concepts_helper import convert_code_list_to_hierarchy


def obtain_reverse_codes(mapped, dst):
    """
    Given the list of desired dst codes and an extensive map src -> dst,
    obtain the list of src codes

    :param mapped: Correspondence between src codes and dst codes [{"o", "to": [{"d", "e"}]}]
    :param dst: Iterable of destination codes
    :return: List of origin codes
    """
    src = set()
    dest_set = set([d.lower() for d in dst])  # Destination categories
    # Obtain origin categories referencing "dest_set" destination categories
    for k in mapped:
        for t in k["to"]:
            if t["d"] and t["d"].lower() in dest_set:
                src.add(k["o"])
    return list(src)  # list(set([k[0].lower() for k in mapped if k[1].lower() in dest_set]))


def pctna(x):
    """
    Aggregation function computing the percentage of NaN values VS total number of elements, in a group "x"
    """
    return 100.0 * np.count_nonzero(np.isnan(x)) / x.size


class ETLExternalDatasetCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        """
        First bring the data considering the filter
        Second, group, third aggregate
        Finally, store the result in State
        """
        issues = []
        # Obtain global variables in state
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)

        # DS Source + DS Name
        source = self._content["dataset_source"]
        dataset_name = self._content["dataset_name"]

        # Result name
        result_name = self._content["result_name"]
        if result_name in datasets or state.get(result_name):
            issues.append((2, "A dataset called '"+result_name+"' is already stored in the registry of datasets"))

        # Dataset metadata
        dims, attrs, meas = obtain_dataset_metadata(dataset_name, source)
        # Obtain filter parameters
        params = create_dictionary()  # Native dimension name to list of values the filter will allow to pass
        joined_dimensions = []
        for dim in self._content["where"]:
            lst = self._content["where"][dim]
            native_dim = None
            if dim.lower() in ["startperiod", "endperiod"]:
                native_dim = dim
                lst = [lst]
            elif dim not in dims:
                # Check if there is a mapping. If so, obtain the native equivalent(s). If not, ERROR
                for m in mappings:
                    if strcmp(mappings[m].destination, dim) and \
                            strcmp(mappings[m].source, source) and \
                            strcmp(mappings[m].dataset, dataset_name) and \
                            mappings[m].origin in dims:
                        joined_dimensions.append(mappings[m].destination)  # Store dimension in the original case
                        native_dim = mappings[m].origin
                        lst = obtain_reverse_codes(mappings[m].map, lst)
                        break
            else:
                # Get the dimension name with the original case
                native_dim = dims[dim].name
            if native_dim:
                if native_dim not in params:
                    f = set()
                    params[native_dim] = f
                else:
                    f = params[native_dim]
                f.update(lst)

        # Convert param contents from set to list
        for p in params:
            params[p] = [i for i in params[p]]

        # Obtain the filtered Dataset <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
        ds = nexinfosys.data_source_manager.get_dataset_filtered(source, dataset_name, params)
        df = ds.data

        # Join with mapped dimensions (augment it)
        # TODO Prepare an "m" containing ALL the mappings affecting "df"
        # TODO df2 = augment_dataframe_with_mapped_columns(df, m, ["value"])
        # TODO Does it allow adding the new column for the dimension, in case it is requested? Probably yes, but test it
        for m in mappings:
            if strcmp(mappings[m].source, source) and \
                    strcmp(mappings[m].dataset, dataset_name) and \
                    mappings[m].origin in dims:
                # TODO Change by many-to-many mapping
                # TODO augment_dataframe_with_mapped_columns(df, maps, measure_columns)
                # Elaborate a many to one mapping
                tmp = []
                for el in mappings[m].map:
                    for to in el["to"]:
                        if to["d"]:
                            tmp.append([el["o"], to["d"]])
                df_dst = pd.DataFrame(tmp, columns=['sou_rce', mappings[m].destination])
                for di in df.columns:
                    if strcmp(mappings[m].origin, di):
                        d = di
                        if not nexinfosys.case_sensitive:
                            df[d+"_l"] = df[d].str.lower()
                            d = d + "_l"
                        break
                df = pd.merge(df, df_dst, how='left', left_on=d, right_on='sou_rce')
                del df['sou_rce']
                if not nexinfosys.case_sensitive:
                    del df[d]

        # Aggregate (If any dimension has been specified)
        if len(self._content["group_by"]) > 0:
            # Column names where data is
            # HACK: for the case where the measure has been named "obs_value", use "value"
            values = [m.lower() if m.lower() != "obs_value" else "value" for m in self._content["measures"]]
            out_names = self._content["measures_as"]
            rows = translate_case(self._content["group_by"], params)  # Group by dimension names
            lcase_rows = [d.lower() for d in rows]
            # Now joined_dimensions
            for d in joined_dimensions:
                if d.lower() in lcase_rows:
                    # Find and replace
                    for i, d2 in enumerate(rows):
                        if strcmp(d, d2):
                            rows[i] = d
                            break

            aggs = []  # Aggregation functions
            agg_names = {}
            for f in self._content["agg_funcs"]:
                if f.lower() in ["avg", "average"]:
                    aggs.append(np.average)
                    agg_names[np.average] = "avg"
                elif f.lower() in ["sum"]:
                    aggs.append(np.sum)
                    agg_names[np.sum] = "sum"
                elif f.lower() in ["count"]:
                    aggs.append(np.size)
                    agg_names[np.size] = "count"
                elif f.lower() in ["sumna"]:
                    aggs.append(np.nansum)
                    agg_names[np.nansum] = "sumna"
                elif f.lower() in ["countav"]:  # countav=="Count Available"
                    aggs.append("count")  # Count number of non-NaN elements
                    agg_names["count"] = "countav"
                elif f.lower() in ["avgav", "avgna"]:  # avgna=="Average without
                    aggs.append(np.nanmean)
                    agg_names[np.nanmean] = "avgna"
                elif f.lower() in ["pctna"]:  # % of NaN vs total elements
                    aggs.append(pctna)
                    agg_names[pctna] = "pctna"

            # Calculate Pivot Table. The columns are a combination of values x aggregation functions
            # For instance, if two values ["v2", "v2"] and two agg. functions ["avg", "sum"] are provided
            # The columns will be: [["average", "v2"], ["average", "v2"], ["sum", "v2"], ["sum", "v2"]]
            try:
                # Check that all "rows" on which pivot table aggregates are present in the input "df"
                # If not either synthesize them (only if there is a single filter value) or remove (if not present
                df_columns_dict = create_dictionary(data={c: c for c in df.columns})
                for r in rows.copy():
                    if r not in df_columns_dict:
                        found = False
                        for k in params:
                            if strcmp(k, r):
                                found = True
                                if len(params[k]) == 1:
                                    df[r] = params[k][0]
                                else:
                                    rows.remove(r)
                                    issues.append((2, "Dimension '" + r + "' removed from the list of dimensions because it is not present in the raw input dataset."))
                                break
                        if not found:
                            rows.remove(r)
                            issues.append((2, "Dimension '" + r + "' removed from the list of dimensions because it is not present in the raw input dataset."))
                # Put proper DIMENSION names
                for ir, r in enumerate(rows):
                    if r in df_columns_dict:
                        rows[ir] = df_columns_dict[r]

                # Create and register Hierarchy objects from origin Dataset dimensions: state, ds
                ds_columns_dict = create_dictionary(data={c.code: c.code for c in ds.dimensions})
                for r in rows:
                    if r in ds_columns_dict:
                        # Create hierarchy local to the dataset
                        for d in ds.dimensions:
                            if strcmp(r, d.code):
                                if d.code_list:
                                    h = convert_code_list_to_hierarchy(d.code_list)
                                    h.name = result_name + "_" + r
                                    glb_idx.put(h.key(), h)
                                    break

                # Pivot table using Group by
                # if True:
                groups = df.groupby(by=rows, as_index=False)  # Split
                d = OrderedDict([])
                lst_names = []
                if len(values) == len(aggs):
                    for i, t in enumerate(zip(values, aggs)):
                        v, agg = t
                        if len(out_names) == len(values):
                            if out_names[i]:
                                lst_names.append(out_names[i])
                            else:
                                lst_names.append(agg_names[agg] + "_" + v)
                        else:
                            lst_names.append(agg_names[agg] + "_" +v)
                        lst = d.get(v, [])
                        lst.append(agg)
                        d[v] = lst
                else:
                    for v in values:
                        lst = d.get(v, [])
                        for agg in aggs:
                            lst.append(agg)
                            lst_names.append(agg_names[agg] + "_" +v)
                        d[v] = lst
                # Print NaN values for each value column
                for v in set(values):
                    cnt = df[v].isnull().sum()
                    logging.debug("NA count for col '"+v+"': "+str(cnt)+" of "+str(df.shape[0]))
                # AGGREGATE !!
                df2 = groups.agg(d)

                # Rename the aggregated columns
                df2.columns = rows + lst_names
                # else:
                #     # Pivot table
                #     df2 = pd.pivot_table(df,
                #                          values=values,
                #                          index=rows,
                #                          aggfunc=[aggs[0]], fill_value=np.NaN, margins=False,
                #                          dropna=True)
                #     # Remove the multiindex in columns
                #     df2.columns = [col[-1] for col in df2.columns.values]
                #     # Remove the index
                #     df2.reset_index(inplace=True)
                # The result, all columns (no index), is stored for later use
                ds.data = df2
            except Exception as e:
                issues.append((3, "There was a problem with the grouping: "+repr(e)))

        # Store the dataset in State
        datasets[result_name] = ds

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Read and check keys validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        return issues





"""
Parsing of fields of the spreadsheet using PyParsing

Ideas for expression rules copied/adapted from:
https://gist.github.com/cynici/5865326

"""
import re
import traceback
from functools import partial

import lxml
import pyparsing
import typing
from pyparsing import (ParserElement, Regex,
                       oneOf, srange, operatorPrecedence, opAssoc,
                       Forward, Regex, Suppress, Literal, Word,
                       Optional, OneOrMore, ZeroOrMore, Or, alphas, alphanums, White,
                       Combine, Group, delimitedList, nums, quotedString, NotAny,
                       Keyword, removeQuotes, CaselessKeyword, OnlyOnce)
from typing import Dict, List

from nexinfosys import ureg
from nexinfosys.command_generators import global_functions, global_functions_extended, extended_dict_of_function_names
from nexinfosys.common.helper import create_dictionary, PartialRetrievalDictionary

# Enable memoizing
# See: https://stackoverflow.com/questions/21370697/pyparsing-performance-and-memory-usage/21371472#21371472
ParserElement.enablePackrat()

# Number
# Variable(key=expression, key=expression)[key=expression, key=expression]
# Hierarchical variable name
# Arithmetic Expression

# FORWARD DECLARATIONS
arith_expression = Forward()
arith_boolean_expression = Forward()
arith_boolean_expression_with_less_tokens = Forward()  # Simpler version of "arith_boolean_expression"
conditions_list = Forward()
expression = Forward()  # Generic expression (boolean, string, numeric)
expression_with_parameters = Forward()  # TODO Parameter value definition. An expression potentially referring to other parameters. Boolean operators. Simulation of IF ELIF ELSE or SWITCH
hierarchy_expression = Forward()
hierarchy_expression_v2 = Forward()
geo_value = Forward()  # TODO Either a Geo or a reference to a Geo
url_parser = Forward()  # TODO
context_query = Forward()  # TODO A way to find a match between pairs of processors. A pair of Processor Selectors
domain_definition = Forward()  # TODO Domain definition. Either a Category Hierarchy name or a numeric interval (with open closed)
parameter_value = Forward()  # TODO Parameter Value. Could be "expression_with_parameters"
indicator_expression = Forward()

# TOKENS

# Separators and operators (arithmetic and boolean)
lparen, rparen, lbracket, rbracket, lcurly, rcurly, dot, equals, hash = map(Literal, "()[]{}.=#")
double_quote = Literal('"')
single_quote = Literal("'")
quote = oneOf('" ''')  # Double quotes, single quotes
signop = oneOf('+ -')
multop = oneOf('* / // %')
plusop = oneOf('+ -')
expop = oneOf('^ **')
comparisonop = oneOf("< <= > >= == != <>")
andop = CaselessKeyword("AND")
orop = CaselessKeyword("OR")
notop = CaselessKeyword("NOT")
processor_factor_separator = Literal(":")
conditions_opening = Literal("?")
conditions_closing = Literal("?")

# Boolean constants
true = Keyword("True")
false = Keyword("False")
# Simple identifier
simple_ident = Word(alphas+"_", alphanums+"_")  # Start in letter and "_", then "_" + letters + numbers
external_ds_name = Word(alphas, alphanums+"-"+"_"+".")  # Dataset names can have
list_simple_ident = delimitedList(simple_ident, ",")

interfaces_list_expression = list_simple_ident
indicators_list_expression = list_simple_ident
attributes_list_expression = list_simple_ident

# Basic data types
positive_int = Word(nums).setParseAction(lambda t: {'type': 'int', 'value': int(t[0])})
positive_float = (Combine(Word(nums) + Optional("." + Word(nums)) + Optional(oneOf("E e") + Optional(oneOf('+ -')) + Word(nums)))
                  ).setParseAction(lambda _s, l, t:
                                   {'type': 'float',
                                    'value': float(t[0])
                                    }
                                   )
signed_float = (Optional(Or([Literal("+"), Literal("-")])("sign")) + positive_float).\
    setParseAction(lambda _s, l, t: {'type': 'float',
                                     'value': t[0]['value'] if isinstance(t[0], dict) else (-t[1]['value'] if t[0]=='-' else t[1]['value'])
                                     }
                   )
boolean = Or([true, false]).setParseAction(lambda t: {'type': 'boolean', 'value': bool(t[0])})

quoted_string = quotedString(r".*")
unquoted_string = Regex(r".*")  # Anything
alphanums_string = Word(alphanums)
code_string = Word(alphanums+"_"+"-")  # For codes in Categories, Code Lists
# Not used
literal_code_string = (Literal('!').suppress()+Optional(simple_ident+dot.suppress())("hierarchy")+code_string("code"))
pair_numbers = signed_float + Literal(",").suppress() + signed_float

# References
code_string_reference = hash.suppress() + code_string + hash.suppress()

# RULES - literal ANY string
string = quotedString.setParseAction(
    lambda t: {'type': 'str', 'value': t[0][1:-1]}
)


# RULES - XQuery/XPath string - used in aggregator expressions potentially involving several processors
# -- NOT USED --
def xquery_validation(s, l, tt):
    """
    Function to elaborate a node for evaluation of processor name (definition of processor name) or
    selection of processors, with variable names and wildcard (..)
    :param s:
    :param l:
    :param tt:
    :return:
    """
    try:
        lxml.etree.XPath(s)
        return {"type": "XQueryString", "query": s}
    except lxml.etree.XPathSyntaxError:
        # TODO Check if it is CSS syntax
        raise Exception("Syntax error validating XQuery/XPath expression")


xpath_string = (Literal("q") + quoted_string).setParseAction(xquery_validation)


# RULES - Processor field string - Name of one Interface, Indicator or Attribute in a Processor
def processor_field_string(s, l, tt):
    """
    :param s:
    :param l:
    :param tt:
    :return:
    """
    # Parse s as simple_ident
    if re.match(r"[a-zA-Z_]\w*", s):
        return {'type': 'ProcessorField', 'value': s}
    else:
        raise Exception("Syntax error validating Processor field string")


processor_field_string = (Literal("p") + quoted_string).setParseAction(processor_field_string)


# RULES - unit name
def parse_action_unit_name(s, l, tt):
    try:
        ureg.parse_expression(s)
        return {"type": "unit_name", "unit": s}
    except:
        raise Exception("Unit name invalid")


unit_name = Regex(r".*").setParseAction(parse_action_unit_name)


# RULES - Simple hierarchical name
#         A literal hierarchical name - A non literal hierarchical name could be "processor_name"
simple_h_name = (simple_ident + ZeroOrMore(dot.suppress() + simple_ident))\
    .setParseAction(lambda _s, l, t: {'type': 'h_var',
                                      'parts': t.asList()
                                      }
                    )

# RULES - simple_hierarchical_name [":" simple_hierarchical_name]
factor_name = (Optional(simple_h_name.setResultsName("processor")) + Optional(Group(processor_factor_separator.suppress() + simple_h_name).setResultsName("factor"))
               ).setParseAction(lambda _s, l, t:
                                {'type': 'pf_name',
                                 'processor': t.processor if t.processor else None,
                                 'factor': t.factor[0] if t.factor else None
                                 }
                                )

# RULES - processor_name
# "h{a.c}ola{b}.sdf{c}",
# "{a.b.c}.h{c}ola{b}.sdf{c}",
# "{a}b",
# "{a}b{c}",
# "aa",
# "aa{b}aa",
# "{a}",
# NOT USED, NOW THERE IS A PREVIOUS MACRO EXPANSION for names like "{a.b}_proc", APPLIED *BEFORE* SYNTAX ANALYSIS
# item = lcurly.suppress() + simple_h_name + rcurly.suppress()
# processor_name_part = Group((item("variable") | Word(alphanums)("literal")) + ZeroOrMore(
#     item("variable") | Word(alphanums + "_")("literal")))


def parse_action_processor_name(s, l, tt, node_type="processor_name"):
    """
    Function to elaborate a node for evaluation of processor name (definition of processor name) or
    selection of processors, with variable names and wildcard (..)
    :param s:
    :param l:
    :param tt:
    :return:
    """
    variables = set()
    parts = []
    expandable = False
    for t in tt:
        st = eval(repr(t))  # Inefficient??
        if st == ".":
            parts.append(("separator", st))
            continue
        elif st == "..":
            parts.append(("separator", st))
            expandable = True
            continue
        classif = st[1]
        for tok in st[0]:
            # Find it in literals or in variables
            if "literal" in classif and tok in classif["literal"]:
                parts.append(("literal", tok))

    expandable = False

    return dict(type=node_type, parts=parts, variables=variables, input=s, expandable=expandable, complex=expandable)


processor_name_part_separator = Literal(".")
processor_name_part = Group(simple_ident("literal"))
# processor_name = simple_h_name TODO Currently this has the syntax of a hierarchical name
processor_name = (processor_name_part("part") + ZeroOrMore(processor_name_part_separator("separator") + processor_name_part("part"))).\
    setParseAction(parse_action_processor_name)

# RULES - processor_names (note the plural)
# "..",
# "..Crop",
# "Farm..Crop",
# "Farm..",
# "..Farm..",
# ".Farm",  # Invalid
# "Farm.",  # Invalid
# "h{c}ola{b}.sdf{c}",
# "{a}.h{c}ola{b}.sdf{c}",
# "{a}b",
# "{a}b{c}",
# "aa",
# "aa{b}aa",
# "{a}",
processor_names_wildcard_separator = Literal("..")
processor_names_pre = (Or([processor_name_part, Literal("*")])("part") +
                       ZeroOrMore((processor_name_part_separator("separator")+processor_name_part("part")) |
                                  (processor_names_wildcard_separator("separator")+Optional(processor_name_part("part")))
                                  )
                       )
processor_names = ((processor_names_wildcard_separator + Optional(processor_names_pre)) | processor_names_pre
                   ).setParseAction(partial(parse_action_processor_name, node_type="processor_names"))


# RULES - context_query
# Right now, context_query would be exactly equal to "processor_names", i.e., a way to specify a set of
# processors (idea proposed by Michele)
context_query << processor_names

# RULES - domain_definition
number_interval = (Or([Literal("["), Literal("(")])("left") + signed_float("number_left") + Literal(",").suppress() + signed_float("number_right") + Or([Literal("]"), Literal(")")])("right"))\
    .setParseAction(lambda _s, l, t: {'type': 'number_interval',
                                      'left': t[0],
                                      'right': t[3],
                                      'number_left': t[1]['value'],
                                      'number_right': t[2]['value'],
                                      })
domain_definition << Or([simple_ident, number_interval])


# RULES - reference
reference = (Optional(lbracket) + simple_ident.setResultsName("ident") + Optional(rbracket)
             ).setParseAction(lambda _s, l, t: {'type': 'reference',
                                                'ref_id': t.ident
                                                }
                              )

bracketed_reference = (lbracket + simple_ident.setResultsName("ident") + rbracket
             ).setParseAction(lambda _s, l, t: {'type': 'reference',
                                                'ref_id': t.ident
                                                }
                              )


# RULES - Processors selector
def processor_selector_validation(s, l, tt):
    """
    Function to elaborate a node for evaluation of processor name (definition of processor name) or
    selection of processors, with variable names and wildcard (..)
    :param s:
    :param l:
    :param tt:
    :return:
    """
    try:
        lxml.etree.XPath(s)
        return s
    except lxml.etree.XPathSyntaxError:
        # TODO Check if it is CSS syntax
        raise Exception("Syntax error validating XPath expression")


processors_selector_expression = Regex(r".*").setParseAction(processor_selector_validation)

# RULES - namespace, parameter list, named parameters list, function call
namespace = simple_ident + Literal("::").suppress()
named_parameter = Group(simple_ident + equals.suppress() + expression).setParseAction(lambda t: {'type': 'named_parameter', 'param': t[0][0], 'value': t[0][1]})
named_parameters_list = delimitedList(named_parameter, ",")
parameters_list = delimitedList(Or([expression, named_parameter]), ",")


def func_call_action(s, l, t):
    if t[0][0] in global_functions:
        return dict(type='function', name=t[0][0], params=t[0][1:])
    else:
        raise Exception(f"Function '{t[0][0]}' not defined")


func_call = Group(simple_ident + lparen.suppress() + parameters_list + rparen.suppress()
                  ).setParseAction(func_call_action)


def func_call_action_extended_list(s, l, t):
    if t[0][0] in extended_dict_of_function_names:
        return dict(type='function', name=t[0][0], params=t[0][1:])
    else:
        raise Exception(f"Function '{t[0][0]}' not defined")


func_call_extended = Group(simple_ident + lparen.suppress() + parameters_list + rparen.suppress()
                           ).setParseAction(func_call_action_extended_list)


# RULES - key-value list
# key "=" value. Key is a simple_ident; "Value" can be an expression referring to parameters
value = Group(arith_boolean_expression).setParseAction(lambda t:
                                                       {'type': 'value',
                                                        'value': t[0]
                                                        }
                                                       )
key_value = Group(simple_ident + equals.suppress() + arith_boolean_expression).setParseAction(lambda t: {'type': 'key_value', 'key': t[0][0], 'value': t[0][1]})
key_value_list = delimitedList(key_value, ",").setParseAction(
    lambda _s, l, t: {'type': 'key_value_list',
                      'parts': {t2["key"]: t2["value"] for t2 in t}
                      }
)


# RULES - dataset variable, hierarchical var name
dataset = (simple_ident.setResultsName("ident") +
           Optional(lparen.suppress() + parameters_list + rparen.suppress()).setResultsName("func_params") +
           lbracket.suppress() + named_parameters_list.setResultsName("slice_params") + rbracket.suppress()
           ).setParseAction(lambda _s, l, t: {'type': 'dataset',
                                              'name': t.ident,
                                              'func_params': t.func_params if t.func_params else None,
                                              'slice_params': t.slice_params,
                                              }
                            )
# [ns::]dataset"."column_name
dataset_with_column = (Optional(namespace).setResultsName("namespace") +
                       Group(Or([simple_ident]) + dot.suppress() + simple_ident).setResultsName("parts")
                       ).setParseAction(lambda _s, l, t:
                                                         {'type': 'h_var',
                                                          'ns': t.namespace[0] if t.namespace else None,
                                                          'parts': t.parts.asList(),
                                                          }
                                        )
# RULES - hierarchical var name
obj_types = Or([simple_ident, func_call, dataset])
# h_name, the most complex NAME, which can be a hierarchical composition of names, function calls and datasets
h_name = (Optional(namespace).setResultsName("namespace") +
          Group(obj_types + ZeroOrMore(dot.suppress() + obj_types)).setResultsName("parts")
          # + Optional(hash + simple_ident).setResultsName("part")
          ).setParseAction(lambda _s, l, t: {'type': 'h_var',
                                             'ns': t.namespace[0] if t.namespace else None,
                                             'parts': t.parts.asList(),
                                             }
                           )

# RULES - Arithmetic expression AND Arithmetic Plus Boolean expression
arith_expression << operatorPrecedence(Or([positive_float, positive_int, string, code_string_reference,
                                       simple_h_name,
                                       func_call]),  # Operand types (REMOVED h_name: no "namespace" and no "datasets")
                                 [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {
                                     'type': 'u'+t.asList()[0][0],
                                     'terms': [0, t.asList()[0][1]],
                                     'ops': ['u'+t.asList()[0][0]]
                                  }),
                                  (expop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'exponentials',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (multop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'multipliers',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'adders',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  ],
                                 lpar=lparen.suppress(),
                                 rpar=rparen.suppress())

# "The" expression, the most complex rule of this file
arith_boolean_expression << operatorPrecedence(Or([positive_float, positive_int,
                                                   string, boolean, code_string_reference,
                                                   bracketed_reference,
                                                   conditions_list,
                                                   simple_h_name,
                                                   func_call]),
                                 [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {
                                     'type': 'u'+t.asList()[0][0],
                                     'terms': [0, t.asList()[0][1]],
                                     'ops': ['u'+t.asList()[0][0]]
                                  }),
                                  (expop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'exponentials',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (multop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'multipliers',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'adders',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (comparisonop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'comparison',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (notop, 1, opAssoc.RIGHT, lambda _s, l, t: {
                                      'type': 'not',
                                      'terms': [0, t.asList()[0][1]],
                                      'ops': ['u'+t.asList()[0][0]]
                                  }),
                                  (andop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'and',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (orop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'or',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  ],
                                 lpar=lparen.suppress(),
                                 rpar=rparen.suppress())

# RULES - Expression varying value depending on conditions
# If a condition omits the condition itself, "True" is assumed, and the result of the expression is returned. So it is recommended for the last in the list
condition = Or([Group(arith_boolean_expression + Literal("->").suppress() + arith_boolean_expression), arith_boolean_expression])
conditions_list << Group(conditions_opening.suppress() + delimitedList(condition, ",") + conditions_closing.suppress()).setParseAction(lambda _s, l, t:
                                                               {
                                                                   'type': 'conditions',
                                                                   'parts': [{"type": "condition", "if": c[0], "then": c[1]} if isinstance(c, list) else
                                                                             {"type": "condition", "if": {'type': 'boolean', 'value': True}, "then": c}
                                                                             for c in t.asList()[0]
                                                                             ]
                                                               })

# RULES - Expression type 2
expression << operatorPrecedence(Or([positive_float, positive_int, string, h_name]),  # Operand types
                                 [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {'type': 'u'+t.asList()[0][0], 'terms': [0, t.asList()[0][1]], 'ops': ['u'+t.asList()[0][0]]}),
                                  (expop, 2, opAssoc.LEFT, lambda _s, l, t: { 'type': 'exponentials', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                  (multop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'multipliers', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                  (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'adders', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                  ],
                                 lpar=lparen.suppress(),
                                 rpar=rparen.suppress())

# RULES - Slightly simpler version of "arith_boolean_expression"
simple_h_name_1 = Group(simple_ident)\
    .setParseAction(lambda _s, l, t: {'type': 'h_var',
                                      'parts': t.asList()[0]
                                      }
                    )

arith_boolean_expression_with_less_tokens << operatorPrecedence(Or([positive_float,
                                                                    positive_int,
                                                                    boolean,
                                                                    string,
                                                                    simple_h_name_1,
                                                                    func_call]),
                                                                [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {
                                     'type': 'u'+t.asList()[0][0],
                                     'terms': [0, t.asList()[0][1]],
                                     'ops': ['u'+t.asList()[0][0]]
                                  }),
                                  (expop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'exponentials',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (multop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'multipliers',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'adders',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (comparisonop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'comparison',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (notop, 1, opAssoc.RIGHT, lambda _s, l, t: {
                                      'type': 'not',
                                      'terms': [0, t.asList()[0][1]],
                                      'ops': ['u'+t.asList()[0][0]]
                                  }),
                                  (andop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'and',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (orop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'or',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  ],
                                                                lpar=lparen.suppress(),
                                                                rpar=rparen.suppress())

# RULES - Expression type 2. Can mention only parameters and numbers
# (for parameters, namespaces are allowed, and also hierarchical naming)
# TODO Check if it can be evaluated with "ast_evaluator"
expression_with_parameters = arith_boolean_expression
# expression_with_parameters << operatorPrecedence(Or([positive_float, positive_int, simple_h_name]),  # Operand types
#                                                  [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {'type': 'u'+t.asList()[0][0], 'terms': [0, t.asList()[0][1]], 'ops': ['u'+t.asList()[0][0]]}),
#                                                   (multop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'multipliers', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
#                                                   (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'adders', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
#                                                   ],
#                                                  lpar=lparen.suppress(),
#                                                  rpar=rparen.suppress())

expression_with_parameters_or_list_simple_ident = Or([arith_boolean_expression, list_simple_ident])

# RULES: Expression type 3. For hierarchies. Can mention only simple identifiers (codes) and numbers
hierarchy_expression << operatorPrecedence(Or([positive_float, positive_int, simple_ident]),  # Operand types
                                           [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {'type': 'u'+t.asList()[0][0], 'terms': [0, t.asList()[0][1]], 'ops': ['u'+t.asList()[0][0]]}),
                                            (expop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'exponentials', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                            (multop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'multipliers', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                            (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'adders', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                            ],
                                           lpar=lparen.suppress(),

                                           rpar=rparen.suppress())

# RULES: Expression type 4. For indicators. Can mention only numbers and core concepts
indicator_expression << operatorPrecedence(Or([positive_float, positive_int,
                                               string, boolean,
                                               conditions_list,
                                               simple_h_name,
                                               func_call_extended]),  # Operand types
                                 [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {
                                     'type': 'u'+t.asList()[0][0],
                                     'terms': [0, t.asList()[0][1]],
                                     'ops': ['u'+t.asList()[0][0]]
                                  }),
                                  (expop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'exponentials',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (multop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'multipliers',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'adders',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (comparisonop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'comparison',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (notop, 1, opAssoc.RIGHT, lambda _s, l, t: {
                                      'type': 'not',
                                      'terms': [0, t.asList()[0][1]],
                                      'ops': ['u'+t.asList()[0][0]]
                                  }),
                                  (andop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'and',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  (orop, 2, opAssoc.LEFT, lambda _s, l, t: {
                                      'type': 'or',
                                      'terms': t.asList()[0][0::2],
                                      'ops': t.asList()[0][1::2]
                                  }),
                                  ],
                                 lpar=lparen.suppress(),
                                 rpar=rparen.suppress())

# [expression2 (previously parsed)] [relation_operator] processor_or_factor_name
relation_operators = Or([Literal('|'),  # Part-of
                         Literal('>'), Literal('<'),  # Directed flow
                         Literal('<>'), Literal('><'),  # Undirected flow
                         Literal('||')  # Upscale (implies part-of also)
                         ]
                        )
relation_expression = (Optional(arith_expression).setResultsName("weight") +
                       Optional(relation_operators).setResultsName("relation_type") +
                       factor_name.setResultsName("destination")
                       ).setParseAction(lambda _s, l, t: {'type': 'relation',
                                                          'name': t.destination,
                                                          'relation_type': t.relation_type,
                                                          'weight': t.weight
                                                          }
                                        )

# RULES: Expression type 5. For hierarchies (version 2). Can mention code_string (for codes), parameters and numbers
hierarchy_expression_v2 << operatorPrecedence(Or([positive_float, positive_int, code_string, named_parameter]),  # Operand types
                                              [(signop, 1, opAssoc.RIGHT, lambda _s, l, t: {'type': 'u'+t.asList()[0][0], 'terms': [0, t.asList()[0][1]], 'ops': ['u'+t.asList()[0][0]]}),
                                            (multop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'multipliers', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                            (plusop, 2, opAssoc.LEFT, lambda _s, l, t: {'type': 'adders', 'terms': t.asList()[0][0::2], 'ops': t.asList()[0][1::2]}),
                                            ],
                                              lpar=lparen.suppress(),
                                              rpar=rparen.suppress())

# RULES: Level name
level_name = (simple_ident + Optional(plusop + positive_int)
              ).setParseAction(lambda t: {'type': 'level',
                                          'domain': t[0],
                                          'level': (t[1][0]+str(t[2]["value"])) if len(t) > 1 and t[1] else None
                                          }
                               )

# RULES: Time expression
# A valid time specification. Possibilities: Year, Month-Year / Year-Month, Time span (two dates)
period_name = Or([Literal("Year"), Literal("Semester"), Literal("Quarter"), Literal("Month")])
four_digits_year = Combine(Word(nums, min=4, max=4)+Optional(Literal(".0")).suppress())
month = Word(nums, min=1, max=2)
year_month_separator = oneOf("- /")
date = Group(Or([four_digits_year.setResultsName("y") +
                 Optional(year_month_separator.suppress()+month.setResultsName("m")),
                 Optional(month.setResultsName("m") + year_month_separator.suppress()) +
                 four_digits_year.setResultsName("y")
                 ]
                )
             )
date_month = Or([four_digits_year + year_month_separator + month, month + year_month_separator + four_digits_year])
two_dates_separator = oneOf("- /")
time_expression = Or([(date + Optional(two_dates_separator.suppress()+date)
                       ).setParseAction(
                                       lambda _s, l, t:
                                       {'type': 'time',
                                        'dates': [{k: int(v) for k, v in d.items()} for d in t]
                                        }
                                       ),
                      period_name.setParseAction(lambda _s, l, t:
                                                 {'type': 'time',
                                                  'period': t[0]})
                      ])

# RULES: "Relative to"
factor_unit = (simple_h_name.setResultsName("factor") + Optional(Regex(".*").setResultsName("unparsed_unit"))).setParseAction(lambda _s, l, t:
                                                                            {'type': 'factor_unit',
                                                                             'factor': t.factor,
                                                                             'unparsed_unit': t.unparsed_unit if t.unparsed_unit else ""})

# #### URL Parser ################################################################

url_chars = alphanums + '-_.~%+'
# "fragment" departs from standard URLs to allow WHITESPACEs (more convenient for users)
fragment = Combine((Suppress('#') + Word(url_chars+" ")))('fragment')
scheme = oneOf('http https ftp file data')('scheme')
host = Combine(delimitedList(Word(url_chars), '.'))('host')
port = Suppress(':') + Word(nums)('port')
user_info = (
Word(url_chars)('username')
  + Suppress(':')
  + Word(url_chars)('password')
  + Suppress('@')
)

query_pair = Group(Word(url_chars) + Suppress('=') + Word(url_chars))
query = Group(Suppress('?') + delimitedList(query_pair, '&'))('query')

path = Combine(
  Suppress('/')
  + OneOrMore(~query + Word(url_chars + '/'))
)('path')

url_parser = (
  scheme.setResultsName("scheme")
  + Suppress('://')
  + Optional(user_info).setResultsName("user_info")
  + Optional(host).setResultsName("host")
  + Optional(port).setResultsName("port")
  + Optional(path).setResultsName("path")
  + Optional(query).setResultsName("query")
  + Optional(fragment).setResultsName("fragment")
).setParseAction(lambda _s, l, t: {'type': 'url',
                                   'scheme': t.scheme,
                                   'user_info': t.user_info,
                                   'host': t.host,
                                   'port': t.port,
                                   'path': t.path,
                                   'query': t.query,
                                   'fragment': t.fragment})


# #################################################################################################################### #


def string_to_ast(rule: ParserElement, input_: str) -> Dict:
    """
    Convert the input string "input_" into an AST, according to "rule"

    :param rule:
    :param input_:
    :return: a dictionary conforming the AST (the format changes from rule to rule)
    """

    def clean_str(us):
        # "En dash"                 character is replaced by minus (-)
        # "Left/Right double quote" character is replaced by double quote (")
        # "Left/Right single quote" character is replaced by single quote (')
        # "€"                       character is replaced by "eur"
        # "$"                       character is replaced by "usd"
        return us.replace(u'\u2013', '-'). \
            replace(u'\u201d', '"').replace(u'\u201c', '"'). \
            replace(u'\u2018', "'").replace(u'\u2019', "'"). \
            replace('€', 'eur'). \
            replace('$', 'usd')

    if rule == unquoted_string:
        return input_
    else:
        res = rule.parseString(clean_str(input_), parseAll=True)
        res = res.asList()[0]
        while isinstance(res, list):
            res = res[0]
        return res


def is_year(s: str) -> bool:
    try:
        four_digits_year.parseString(s, parseAll=True)
    except pyparsing.ParseException:
        return False
    return True


def is_month(s: str) -> bool:
    try:
        date_month.parseString(s, parseAll=True)
    except pyparsing.ParseException:
        return False
    return True


def parse_string_as_simple_ident_list(in_list: str) -> typing.Optional[List[str]]:
    if in_list:
        try:
            res = list_simple_ident.parseString(in_list, parseAll=True)
            if res:
                return res.asList()
        except pyparsing.ParseException:
            pass
    return None


if __name__ == '__main__':
    from nexinfosys.model_services import State
    from dotted.collection import DottedDict

    number_interval_examples = [
        "(1, 2)",
        "[0.1, 1.0]",
        "[-3.14, 1)",
        "(-1, 3]"
    ]

    for e in number_interval_examples:
        try:
            ast = string_to_ast(number_interval, e)
            print(ast)
        except:
            print("Incorrect")

    processor_name_examples = [
        "BFGas_DE_2016",
        "BFGas_DE_2016.b"
    ]
    for e in processor_name_examples:
        try:
            ast = string_to_ast(processor_name, e)
            print(ast)
        except:
            print("Incorrect")

    examples = ["?a > 5 -> 3, a> 10 -> 6, 8?",
                "a * 3 >= 0.3",
                "'Hola'",
                "'Hola' + 'Adios'",
                "Param * 3 >= 0.3",
                "5 * Param1",
                "True",
                "(Param * 3 >= 0.3) AND (Param2 * 2 <= 0.345)",
                "cos(Param*3.1415)",
                "'Hola' + Param1"
    ]
    for e in examples:
        try:
            ast = string_to_ast(arith_boolean_expression, e)
            print(ast)
        except:
            traceback.print_exc()
            print("Incorrect")

    s = "c1 + c30 - c2 - 10"
    res = string_to_ast(hierarchy_expression, s)
    s = "ds.col"
    res = string_to_ast(h_name, s)

    s = State()
    examples = [
        "5-2017",
        "2017-5",
        "2017/5",
        "2017-05 - 2018-01",
        "2017",
        "5-2017 - 2018-1",
        "2017-2018",
        "Year",
        "Month"
    ]
    for example in examples:
        print(example)
        res = string_to_ast(time_expression, example)
        print(res)
        print(f'Is year = {is_year(example)}')
        print(f'Is month = {is_month(example)}')
        print("-------------------")

    for list1 in ["this, is, a,simple_ident , list", "this,is,'NOT', a,simple_ident , list", " word ", "", " , ", None]:
        parsed_list = parse_string_as_simple_ident_list(list1)
        print(f'Parsing list "{list1}": {parsed_list}')

    s.set("HH", DottedDict({"Power": {"p": 34.5, "Price": 2.3}}))
    s.set("EN", DottedDict({"Power": {"Price": 1.5}}))
    s.set("HH", DottedDict({"Power": 25}), "ns2")
    s.set("param1", 0.93)
    s.set("param2", 0.9)
    s.set("param3", 0.96)
    examples = [
        "EN(p1=1.5, p2=2.3)[d1='C11', d2='C21'].v2",  # Simply sliced Variable Dataset (function call)
        "a_function(p1=2, p2='cc', p3=1.3*param3)",
        "-5+4*2",  # Simple expression #1
        "HH",  # Simple name
        "HH.Power.p",  # Hierarchic name
        "5",  # Integer
        "1.5",  # Float
        "1e-10",  # Float scientific notation
        "(5+4)*2",  # Simple expression #2 (parenthesis)
        "3*2/6",  # Simple expression #3 (consecutive operators of the same kind)
        "'hello'",  # String
        "ns2::HH.Power",  # Hierarchic name from another Namespace
        "HH.Power.Price + EN.Power.Price * param1",
        "EN[d1='C11', d2='C21'].d1",  # Simple Dataset slice
        "b.a_function(p1=2, p2='cc', p3=1.3*param3)",
        "b.EN[d1='C11', d2='C21'].d1",  # Hierachical Dataset slice
        "tns::EN(p1=1.5+param2, p2=2.3 * 0.3)[d1='C11', d2='C21'].v2",  # Simply sliced Variable Dataset (function call)
    ]
    # for example in examples:
    #     print(example)
    #     res = string_to_ast(expression, example)
    #     print(res)
    #     issues = []
    #     value = ast_evaluator(res, s, None, issues)
    #     print(str(type(value))+": "+str(value))




import json

from xml.sax.saxutils import escape

from nexinfosys import metadata_fields
from nexinfosys.common.helper import create_dictionary
from nexinfosys.model_services import IExecutableCommand


def generate_dublin_core_xml(content):
    """
    Generate an XML string with a Simple Dublin Core Record from a Case Study Metadata Command Content
    :param content:
    :return:
    """
    controlled = create_dictionary()
    for t in metadata_fields:
        controlled[t[4]] = t

    s = """<?xml version="1.0"?>
<caseStudyMetadata xmlns="http://magic-nexus.org/dmp/" xmlns:dc="http://purl.org/dc/elements/1.1/">
"""
    for key in content:
        k = controlled[key][1]
        if k:
            for l in content[key]:
                s += "    <dc:" + k + ">" + escape(str(l)) + "</dc:" + k + ">\n"

    s += "</caseStudyMetadata>\n"

    return s


class MetadataCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._metadata_dictionary = {}

    def execute(self, state: "State"):
        """ The execution creates an instance of a Metadata object, and assigns the name "metadata" to the variable,
            inserting it into "State" 
        """
        issues = []
        cs = state.get("_case_study")
        cs_version = state.get("_case_study_version")
        state.set("_metadata", self._metadata_dictionary)
        if cs:
            # Modify case study attributes
            cs_version.name = ""
            if "case_study_name" in self._metadata_dictionary and self._metadata_dictionary["case_study_name"]:
                cs_version.name += "- ".join(self._metadata_dictionary["case_study_name"])
            if "title" in self._metadata_dictionary and self._metadata_dictionary["title"]:
                if cs_version.name:
                    cs_version.name += "; "
                cs_version.name += "- ".join(self._metadata_dictionary["title"])
            if "doi" in self._metadata_dictionary and self._metadata_dictionary["doi"]:
                cs.oid = self._metadata_dictionary["doi"][0]
            if "description" in self._metadata_dictionary and self._metadata_dictionary["description"]:
                cs.description = '; '.join(self._metadata_dictionary["description"])
            if "dimensions" in self._metadata_dictionary and self._metadata_dictionary["dimensions"]:
                cs.areas = ""
                lst = [i.lower() for i in self._metadata_dictionary["dimensions"]]
                if "water" in lst:
                    cs.areas += "W"
                if "energy" in lst:
                    cs.areas += "E"
                if "food" in lst:
                    cs.areas += "F"
                if "land" in lst:
                    cs.areas += "L"
                if "climate" in lst:
                    cs.areas += "C"
            if "geographical_level" in self._metadata_dictionary and self._metadata_dictionary["geographical_level"]:
                cs.geographic_level = ""
                lst = [i.lower() for i in self._metadata_dictionary["geographical_level"]]
                if "local" in lst:
                    cs.geographic_level += "L"
                if "regional" in lst or "region" in lst:
                    cs.geographic_level += "R"
                if "country" in lst:
                    cs.geographic_level += "C"
                if "europe" in lst:
                    cs.geographic_level += "E"
                if "sector" in lst or "sectoral" in lst:
                    cs.geographic_level += "S"
            if "restriction_level" in self._metadata_dictionary and self._metadata_dictionary["restriction_level"]:
                cs.restriction_level = ""
                lst = [i.lower() for i in self._metadata_dictionary["restriction_level"]]
                if "internal" in lst:
                    cs.restriction_level += "I"
                if "confidential" in lst:
                    cs.restriction_level += "C"
                if "public" in lst:
                    cs.restriction_level += "P"
            if "version" in self._metadata_dictionary and self._metadata_dictionary["version"]:
                cs.version = str(self._metadata_dictionary["version"])

            if "case_study_code" not in self._metadata_dictionary:
                # Generate internal code according to DMP
                correlative = 1  # TODO Obtain unique -MAGIC project level- correlative number
                cs.internal_code = "CS"+str(correlative)+"_"+cs.geographic_level+"_"+cs.areas+"_"+cs.restriction_level+"-"+cs.version
                self._metadata_dictionary["case_study_code"] = cs.internal_code
            else:
                cs.internal_code = self._metadata_dictionary["case_study_code"][0]

        return issues, None  # Issues, output

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._metadata_dictionary

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, (dict, list)):
            self._metadata_dictionary = json_input
        else:
            self._metadata_dictionary = json.loads(json_input)
        return issues


import getpass
import logging
import os
import tempfile
import zipfile
from typing import List
import pandas as pd

from nexinfosys import get_global_configuration_variable

pd.core.common.is_list_like = pd.api.types.is_list_like
import pandas_datareader.data as web
import numpy as np
import requests
import datetime
from io import StringIO

from nexinfosys.common.helper import Memoize2, import_names
from nexinfosys.ie_imports.data_source_manager import IDataSourceManager, filter_dataset_into_dataframe, \
    get_dataset_structure
from nexinfosys.models.statistical_datasets import DataSource, Database, Dataset, Dimension, CodeList, CodeImmutable

base_url = "http://ec.europa.eu/agriculture/rica/database/reports/"
measures_file = "http://ec.europa.eu/agriculture/rica/database/help/infometa.csv"

dimensions = {"Year": ("YEAR", "Y", "Year"), # Dimension name: (Dimension name in files, Dimension name in dataset code)
              "Country": ("COUNTRY", "C", "Country"),
              "Region": ("REGION", "R", "Region"),
              "TypeOfFarming8": ("TF8", "F8", "Type of farming 8 FADN"),
              "TypeOfFarming14": ("TF14", "F14", "Type of farming 14 FADN"),
              "TypeOfFarmingGen1": ("TF_GEN1", "FG", "Type of farming from EU Typology Gen 1"),
              "TypeOfFarmingPrin2": ("TF_PRIN2", "FP", "Type of farming from EU Typology Prin 2"),
              "TypeOfFarmingSubP4": ("TF_SUBP4", "FS", "Type of farming from EU Typology Sub P 4"),
              "EconomicSize6": ("SIZ6", "S6", "Economic Size 6 FADN"),
              "EconomicSizeEUTypology": ("SIZC", "SEU", "Economic Size EU Typology"),
              "OrganicProduction": ("ORGANIC", "ORG", "Organic Production categorization"),
              "ANC3": ("ANC3", "ANC", "ANC3"),
              "LFA": ("LFA", "LFA", "LFA")
              }

measures = {}  # Download and parse from  file at http://ec.europa.eu/agriculture/rica/database/help/infometa.csv

# Label and Code are computed from the dimension, plus
# the list is replicated for the combination SO/SGO and MEAN/MEDIAN, changing the label and the code used
datasets = [dict(dimensions=["Year", "Country"]),
            dict(dimensions=["Year", "Country", "TypeOfFarming8"]),
            dict(dimensions=["Year", "Country", "TypeOfFarming14"]),
            dict(dimensions=["Year", "Country", "TypeOfFarmingGen1", "TypeOfFarmingPrin2", "TypeOfFarmingSubP4"]),
            dict(dimensions=["Year", "Country", "EconomicSize6"]),
            dict(dimensions=["Year", "Country", "EconomicSizeEUTypology"]),
            dict(dimensions=["Year", "Country", "ANC3"]),
            dict(dimensions=["Year", "Country", "LFA"]),
            dict(dimensions=["Year", "Country", "OrganicProduction"]),
            dict(dimensions=["Year", "Country", "EconomicSize6", "TypeOfFarming8"]),
            dict(dimensions=["Year", "Country", "EconomicSize6", "TypeOfFarming14"]),
            dict(dimensions=["Year", "Country", "Region"]),
            dict(dimensions=["Year", "Country", "Region", "TypeOfFarming8"]),
            dict(dimensions=["Year", "Country", "Region", "TypeOfFarming14"]),
            dict(dimensions=["Year", "Country", "Region", "EconomicSize6"]),
            dict(dimensions=["Year", "Country", "Region", "EconomicSizeEUTypology"]),
            dict(dimensions=["Year", "Country", "Region", "EconomicSize6", "TypeOfFarming8"])
            ]

versions = sorted(["20120827", "20121212", "20130313", "20130521", "20130605", "20131108", "20131211",
                   "20140312", "20140325", "20140720", "20140917", "20141107",
                   "20150107", "20150504", "20160107", "20160217", "20170126", "20170129", "20170518",
                   "20170608", "20170717", "20170912", "20171113", "20171114", "20180120", "20180518", "20180610"
                   ])
# Number of datasets
# 15 * 2 (SO, SGM) * 2 (Media, Mediana) * number of versions


def generate_datasets(seed_datasets):
    ds_out = []
    for ds in seed_datasets:
        label = " - ".join(ds["dimensions"])
        fname = ".".join([dimensions[t][0] for t in ds["dimensions"]])
        code = ".".join([dimensions[t][1] for t in ds["dimensions"]])

        ds_out.append(dict(dimensions=ds["dimensions"], fname=fname, code=code+".AVG.SO", label=label + " - SO"))
        ds_out.append(dict(dimensions=ds["dimensions"], fname=fname+".MED", code=code+".MED.SO", label=label + " - SO (Mediane)"))
        ds_out.append(dict(dimensions=ds["dimensions"], fname=fname, code=code+".AVG.SGM", label=label + " - SGM"))
        ds_out.append(dict(dimensions=ds["dimensions"], fname=fname+".MED", code=code+".MED.SGM", label=label + " - SGM (Mediane)"))

    return ds_out


def get_fadn_directory():
    # EuroStat datasets
    if get_global_configuration_variable('FADN_FILES_LOCATION'):
        dir_name = get_global_configuration_variable('FADN_FILES_LOCATION')
        logging.debug(f"USER: {getpass.getuser()}, creating dir for FADN files")
    else:
        dir_name = tempfile.gettempdir() + "/fadn_datasets"

    if not os.path.isdir(dir_name):
        os.makedirs(dir_name)

    return dir_name


def download(url, file):
    logging.debug("download URL: "+url)
    logging.debug("file: "+file)
    if not os.path.isfile(file):
        r = requests.get(url, stream=True)
        if r.status_code == 200:
            # http://stackoverflow.com/questions/15352668/download-and-decompress-gzipped-file-in-memory
            with open(file, 'wb') as f:
                for chunk in r.iter_content(chunk_size=1024):
                    if chunk:  # filter out keep-alive new chunks
                        f.write(chunk)


def load_measures_dictionary(directory):
    fname = directory+"/measures.csv"
    if not os.path.isfile(fname):
        download(measures_file, fname)
    if not os.path.isfile(fname):
        raise Exception("Could not obtain measures dictionary file")
    # Read the file into a DataFrame
    with open(fname, "rb") as f:
        s = f.read().decode("ISO-8859-1")
    df = pd.read_csv(StringIO(s.replace("\"", "")), sep="^")
    d = {}
    en_pos = 1
    formula_pos = df.shape[1]-1
    for r in range(df.shape[0]):
        unit1 = None
        unit2 = None
        formula = df.iloc[r, formula_pos]
        if "pre-2014" in formula:
            s = formula[formula.find("pre-2014")+9:]
            unit1 = s[s.find("(")+1:s.find(")")]

        if "post-2013" in formula:
            s = formula[formula.find("post-2013")+10:]
            unit2 = s[s.find("(")+1:s.find(")")]

        add = True
        if unit1 and unit2:
            if unit1 != unit2:
                logging.debug("Different: "+df.iloc[r, 0]+"; "+formula)
                add = False

        if add:
            d[df.iloc[r, 0]] = (df.iloc[r, en_pos], formula)

    return d


def load_dataset(code, date, ds_lst, directory, base_url=base_url):
    """
    Loads a dataset into a DataFrame
    If the dataset is present, it decompresses it in memory to obtain one of the four datasets per file
    If the dataset is not downloaded, downloads it and decompresses into the corresponding version directory
    :param code:
    :param date:
    :param ds_lst: list of FADN datasets
    :param directory:
    :param base_url:
    :return:
    """

    def get_version_date():
        # Version date from date parameter
        vdate = None
        if not date:
            vdate = versions[-1]
        else:
            if isinstance(date, datetime.datetime):
                sdate = date.strftime("%Y%m%d")
            elif isinstance(date, str):
                sdate = date
            # Find previous nearest
            for v in reversed(versions):
                if v < sdate:
                    version_date = v
                    break
        return vdate

    version_date = get_version_date()

    # Dataset dictionary
    d = None
    for ds in ds_lst:
        if ds["code"] == code:
            d = ds
            break

    # Dir name
    dname = directory + os.sep + version_date
    # File name
    fname = dname + os.sep + d["fname"] + ".zip"

    # Check if the file exists
    if not os.path.isfile(fname):
        # If it is the latest version
        if version_date == versions[-1] and "MED" not in code:
            os.makedirs(dname)
            # Download directly the file
            download(base_url+d["fname"]+".zip", fname)
        else:
            # Download the full package
            zip_name = directory+f"{os.sep}fadn"+version_date+".zip"
            download(base_url+f"{os.sep}archives{os.sep}fadn"+version_date+".zip", zip_name)
            # Create directory
            target_name = directory + os.sep + version_date
            os.makedirs(target_name)
            # Uncompress it into the version directory
            z = zipfile.ZipFile(zip_name)
            z.extractall(target_name)
            z.close()

    # At this point the file should exist. If not, raise an Exception
    if not os.path.isfile(fname):
        raise Exception("Could not obtain the dataset "+code+" for processing")

    # Obtain the file from the ZIP file
    z = zipfile.ZipFile(fname)
    if code.endswith(".SO"):
        path = "SO"
    else:
        path = "SGM"
    b = z.open(path + os.sep + d["fname"] + ".csv").read().decode("utf-8")
    df = pd.read_csv(StringIO(b), sep=";")

    # Split dimension columns in two (except Year)
    # Also, remove rows where a dimension is "NaN"
    for dim in d["dimensions"]:
        if dim == "Year":
            continue
        col = dimensions[dim][0]
        # Remove rows where
        df = df[~df[col].isnull()]
        # Do the split
        tmp = df[col].str.split(")", 1).str
        df[col] = tmp[0].str[1:]
        df[col+"_DESC"] = tmp[1].str[1:]

    return df


class FADN(IDataSourceManager):
    def __init__(self, metadata_session_factory, data_engine):
        self._datasets = generate_datasets(datasets)
        self._base_directory = get_fadn_directory()
        self._measures = load_measures_dictionary(self._base_directory)
        self._metadata_session_factory = metadata_session_factory
        self._data_engine = data_engine

    def get_name(self) -> str:
        """ Source name """
        return self.get_datasource().name

    def get_datasource(self) -> DataSource:
        """ Data source """
        src = DataSource()
        src.name = "FADN"
        src.description = "The Farm Accountancy Data Network (FADN) is an instrument for evaluating the income of agricultural holdings and the impacts of the Common Agricultural Policy"
        return src

    def get_databases(self) -> List[Database]:
        """ List of databases in the data source """
        db = Database()
        db.code = ""
        db.description = "FADN provides all Datasets in a single database"
        return [db]

    @Memoize2
    def get_datasets(self, database=None) -> list:
        """ List of datasets in a database, or in all the datasource (if database==None)
            Return a list of tuples (database, dataset)
        """

        lst = []
        for d in self._datasets:
            lst.append((d["code"], d["label"]))

        return lst

    def get_dataset_structure(self, database, dataset) -> Dataset:
        """ Obtain the structure of a dataset: concepts, dimensions, attributes and measures """

        return get_dataset_structure(self._metadata_session_factory, self, dataset)

    def etl_full_database(self, database=None, update=False):
        """ If bulk download is supported, refresh full database """
        pass

    def etl_dataset(self, dataset, update=False) -> Dataset:
        """
        Download a FADN bulk file

        :param url:
        :param local_filename:
        :param update:
        :return: String with full file name
        """

        # Obtain the dictionary
        d = None
        for ds in self._datasets:
            if ds["code"] == dataset:
                d = ds
                break

        # Read the full file, but only to extract the code lists
        df = load_dataset(dataset, None, self._datasets, self._base_directory)

        # ---------------------------------------------------------
        #     Elaborate the dataset, with dimensions and measures
        # ---------------------------------------------------------
        ds = Dataset()
        ds.code = dataset
        ds.description = d["label"]
        ds.attributes = {}  # Dataset level attributes? (encode them using a dictionary)
        ds.metadata = None  # Metadata for the dataset SDMX (flow, date of production, etc.)
        ds.database = self.get_databases()[0]  # Assign in the calling function

        # Add Dimensions
        for dim in d["dimensions"]:
            dd = Dimension()
            dd.code = dimensions[dim][0]
            dd.description = dimensions[dim][2]
            dd.attributes = None
            dd.is_time = dim == "Year"
            dd.is_measure = False
            if dim == "Year":
                cl = sorted(list([str(c) for c in df[dd.code].unique() if c != np.nan]))
                dd.code_list = CodeList.construct(dim, dim, [""], [CodeImmutable(c, c, "", []) for c in cl])
            else:
                df_tmp = df[[dd.code, dd.code + "_DESC"]].drop_duplicates()
                cl = sorted(list([str(c) for c in df_tmp[dd.code].unique() if c != np.nan]))
                df_tmp.set_index(dd.code, inplace=True)
                dd.code_list = CodeList.construct(dim, dim, [""], [CodeImmutable(c, df_tmp.loc[c].values[0], "", []) for c in cl])

            dd.dataset = ds

        # Add Measures
        for m in df.columns:
            if m.endswith("_Q50"):
                t = m[:-len("_Q50")]
            else:
                t = m
            if t in self._measures:
                dd = Dimension()
                dd.code = m
                dd.description = self._measures[t][0]
                dd.attributes = None
                dd.is_time = False
                dd.is_measure = True
                dd.code_list = None
                dd.dataset = ds

        return ds

    def get_dataset_filtered(self, dataset: str, dataset_params: dict) -> Dataset:
        """ This method has to consider the last dataset download, to re"""

        # Read dataset structure
        ds = self.get_dataset_structure(None, dataset)

        df = load_dataset(dataset, None, self._datasets, self._base_directory)

        # Obtain dataset dictionary
        d = None
        for dd in self._datasets:
            if dd["code"] == dataset:
                d = dd
                break

        # Remove descriptions
        # Lower case contents of dimension columns
        lst_idx_cols = []
        for dim in d["dimensions"]:
            col = dimensions[dim][0]
            lst_idx_cols.append(col.lower())
            if col.lower() != "year":
                del df[col+"_DESC"]
                df[col] = df[col].str.lower()

        # Lower case column names
        df.columns = [c.lower() for c in df.columns]
        df.set_index(lst_idx_cols, inplace=True)

        # Filter it using generic Pandas filtering capabilities
        if dataset_params["StartPeriod"] and dataset_params["EndPeriod"]:
            years = [str(y) for y in range(int(dataset_params["StartPeriod"][0]), int(dataset_params["EndPeriod"][0])+1)]
            dataset_params["year"] = years
            del dataset_params["StartPeriod"]
            del dataset_params["EndPeriod"]
        ds.data = filter_dataset_into_dataframe(df, dataset_params, dataset)

        return ds

    def get_refresh_policy(self):  # Refresh frequency for list of databases, list of datasets, and dataset
        pass


if __name__ == '__main__':
    datasets = generate_datasets(datasets)
    d = load_measures_dictionary("/home/rnebot/Downloads/fadn")
    df = load_dataset("Y.C.MED.SO", None, datasets, "/home/rnebot/Downloads/fadn")
    print(df)



# -*- coding: utf-8 -*-
import ast
import base64
import collections
import io
import itertools
import json
import logging
import mimetypes
import re
import tempfile
import urllib
import urllib.request
import uuid
from enum import Enum
from functools import partial, reduce
from operator import add, mul, sub, truediv
from typing import IO, List, Tuple, Dict, Any, Optional, Iterable, Callable, TypeVar, Type, Union, SupportsFloat
from urllib.parse import urlparse
from uuid import UUID

import jsonpickle
import numpy as np
import pandas as pd
import pycurl
from multidict import MultiDict, CIMultiDict
from pandas import DataFrame

import nexinfosys
from nexinfosys import case_sensitive, SDMXConcept, get_global_configuration_variable
from nexinfosys.common.decorators import deprecated
from nexinfosys.ie_imports.google_drive import download_xlsx_file_id
from nexinfosys.models import log_level

logger = logging.getLogger(__name__)
logger.setLevel(log_level)


def append_line(str, file=None):
    if not file:
        file = "/app/borrame.txt"
    f = open(file, "a+")
    f.write(str+"\n")
    f.close()


# #####################################################################################################################
# >>>> JSON FUNCTIONS <<<<
# #####################################################################################################################

def _json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""
    from datetime import datetime
    from nexinfosys.command_generators import Issue  # Imported here to break circularity of imports

    if isinstance(obj, datetime):
        serial = obj.isoformat()
        return serial
    elif isinstance(obj, CaseInsensitiveDict):
        return str(obj)
    elif isinstance(obj, np.int64):
        return int(obj)
    elif isinstance(obj, Issue):
        return obj.__repr__()
    raise TypeError(f"Type {type(obj)} not serializable")


JSON_INDENT = 4
ENSURE_ASCII = False


def generate_json(o):
    return json.dumps(o,
                      default=_json_serial,
                      sort_keys=True,
                      indent=JSON_INDENT,
                      ensure_ascii=ENSURE_ASCII,
                      separators=(',', ': ')
                      ) if o else None


class Encodable:
    """
    Abstract class with the method encode() that should be implemented by a subclass to be encoded into JSON
    using the json.dumps() method together with the option cls=CustomEncoder.
    """
    def encode(self) -> Dict[str, Any]:
        raise NotImplementedError("users must define encode() to use this base class")

    @staticmethod
    def parents_encode(obj: "Encodable", cls: type) -> Dict[str, Any]:
        """
        Get the state of all "cls" parent classes for the selected instance "obj"
        :param obj: The instance. Use "self".
        :param cls: The base class which parents we want to get. Use "__class__".
        :return: A dictionary with the state of the instance "obj" for all inherited classes.

        """
        d = {}
        for parent in cls.__bases__:
            if issubclass(parent, Encodable) and parent is not Encodable:
                d.update(parent.encode(obj))
        return d


class CustomEncoder(json.JSONEncoder):
    """
    Encoding class used by json.dumps(). It should be passed as the "cls" argument.
    Example: print(json.dumps({'A': 2, 'b': 4}), cls=CustomEncoder)
    """
    def default(self, obj):
        # Does the object implement its own encoder?
        if isinstance(obj, Encodable):
            return obj.encode()

        # Use the default encoder for handled types
        if isinstance(obj, (list, dict, str, int, float, bool, type(None))):
            return json.JSONEncoder.default(self, obj)

        # For other unhandled types, like set, use universal json encoder "jsonpickle"
        return jsonpickle.encode(obj, unpicklable=False)


# #####################################################################################################################
# >>>> CASE SeNsItIvE or INSENSITIVE names (flows, funds, processors, ...) <<<<
# #####################################################################################################################
# from nexinfosys.models.musiasem_concepts import Taxon  IMPORT LOOP !!!!! AVOID !!!!


class CaseInsensitiveDict(collections.MutableMapping, Encodable):
    """
    A dictionary with case insensitive Keys.
    Prepared also to support TUPLES as keys, required because compound keys are required
    """
    def __init__(self, data=None, **kwargs):
        from collections import OrderedDict
        self._store = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def encode(self):
        return self.get_data()

    def get_original_data(self):
        return {casedkey: mappedvalue for casedkey, mappedvalue in self._store.values()}

    def get_data(self):
        return {key: self._store[key][1] for key in self._store}

    def __setitem__(self, key, value):
        # Use the lowercased key for lookups, but store the actual
        # key alongside the value.
        if not isinstance(key, tuple):
            self._store[key.lower()] = (key, value)
        else:
            self._store[tuple([k.lower() for k in key])] = (key, value)

    def __getitem__(self, key):
        if not isinstance(key, tuple):
            return self._store[key.lower()][1]
        else:
            return self._store[tuple([k.lower() for k in key])][1]

    def __delitem__(self, key):
        if not isinstance(key, tuple):
            del self._store[key.lower()]
        else:
            del self._store[tuple([k.lower() for k in key])]

    def __iter__(self):
        return (casedkey for casedkey, mappedvalue in self._store.values())

    def __len__(self):
        return len(self._store)

    def lower_items(self):
        """Like iteritems(), but with all lowercase keys."""
        return (
            (lowerkey, keyval[1])
            for (lowerkey, keyval)
            in self._store.items()
        )

    def __contains__(self, key):  # "in" operator to check if the key is present in the dictionary
        if not isinstance(key, tuple):
            return key.lower() in self._store
        else:
            return tuple([k.lower() for k in key]) in self._store

    def __eq__(self, other):
        if isinstance(other, collections.Mapping):
            other = CaseInsensitiveDict(other)
        else:
            return NotImplemented
        # Compare insensitively
        return dict(self.lower_items()) == dict(other.lower_items())

    # Copy is required
    def copy(self):
        return CaseInsensitiveDict(self._store.values())

    def __repr__(self):
        return str(dict(self.items()))


def create_dictionary(case_sens=case_sensitive, multi_dict=False, data=dict()):
    """
    Factory to create dictionaries

    :param case_sens: True to create a case sensitive dictionary, False to create a case insensitive one
    :param multi_dict: True to create a "MultiDict", capable of storing several values
    :param data: Dictionary with which the new dictionary is initialized
    :return:
    """

    if not multi_dict:
        if case_sens:
            tmp = {}
            tmp.update(data)
            return tmp  # Normal, "native" dictionary
        else:
            return CaseInsensitiveDict(data)
    else:
        if case_sens:
            return MultiDict(data)
        else:
            return CIMultiDict(data)


def strcmp(s1, s2):
    """
    Compare two strings for equality or not, considering a flag for case sensitiveness or not

    It also removes leading and trailing whitespace from both strings, so it is not sensitive to this possible
    difference, which can be a source of problems

    :param s1:
    :param s2:
    :return:
    """
    # Handling empty or None strings
    if not s1:
        return True if not s2 else False
    if not s2:
        return False

    if case_sensitive:
        return s1.strip() == s2.strip()
    else:
        return s1.strip().lower() == s2.strip().lower()


def istr(s1: str) -> str:
    """ Return a lowercase version of a string if program works ignoring case sensitiveness """
    if case_sensitive:
        return s1
    else:
        return s1.lower()

# #####################################################################################################################
# >>>> DYNAMIC IMPORT <<<<
# #####################################################################################################################


def import_names(package, names):
    """
    Dynamic import of a list of names from a module

    :param package: String with the name of the package
    :param names: Name or list of names, string or list of strings with the objects inside the package
    :return: The list (or not) of objects under those names
    """
    if not isinstance(names, list):
        names = [names]
        not_list = True
    else:
        not_list = False

    try:
        tmp = __import__(package, fromlist=names)
    except:
        tmp = None

    if tmp:
        tmp2 = [getattr(tmp, name) for name in names]
        if not_list:
            tmp2 = tmp2[0]
        return tmp2
    else:
        return None

# #####################################################################################################################
# >>>> KEY -> VALUE STORE, WITH PARTIAL KEY INDEXATION <<<<
# #####################################################################################################################


@deprecated  # DEPRECATED!!!: Use PartialRetrievalDictionary
class PartialRetrievalDictionary2:
    """
    Implementation using pd.DataFrame, very slow!!!

    The key is a dictionary, the value an object. Allows partial search.

    >> IF "case_sensitive==False" -> VALUES are CASE INSENSITIVE <<<<<<<<<<<<<<<<<

        It is prepared to store different key compositions, with a pd.DataFrame per key
        When retrieving (get), it can match several of these, so it can return results from different pd.DataFrames

        pd.DataFrame with MultiIndex:

import pandas as pd
df = pd.DataFrame(columns=["a", "b", "c"])  # Empty DataFrame
df.set_index(["a", "b"], inplace=True)  # First columns are the MultiIndex
df.loc[("h", "i"), "c"] = "hi"  # Insert values in two cells
df.loc[("h", "j"), "c"] = "hj"
df.loc[(slice(None), "j"), "c"]  # Retrieve using Partial Key
df.loc[("h", slice(None)), "c"]
    """
    def __init__(self):
        self._dfs = {}  # Dict from sorted key-dictionary-keys (the keys in the key dictionary) to DataFrame
        self._df_sorted = {}  # A Dict telling if each pd.DataFrame is sorted
        self._key_lst = []  # List of sets of keys, used in "get" when matching partial keys

    def put(self, key, value):
        """
        Store a value using a dictionary key which can have None values

        :param key:
        :param value:
        :return:
        """
        # Sort keys
        keys = [k for k in key]
        s_tuple = tuple(sorted(keys))
        if case_sensitive:
            df_key = tuple([key[k] if key[k] is not None else slice(None) for k in s_tuple])
        else:
            df_key = tuple([(key[k] if k.startswith("__") else key[k].lower()) if key[k] is not None else slice(None) for k in s_tuple])

        if s_tuple not in self._dfs:
            # Append to list of sets of keys
            self._key_lst.append(set(keys))
            # Add New DataFrame to the dictionary of pd.DataFrame
            cols = [s for s in s_tuple]
            cols.append("value")
            df = pd.DataFrame(columns=cols)
            self._dfs[s_tuple] = df
            df.set_index([s for s in s_tuple], inplace=True)
        else:
            # Use existing pd.DataFrame
            df = self._dfs[s_tuple]

        # Do the insertion into the pd.DataFrame
        df.loc[df_key, "value"] = value

        # Flag the pd.DataFrame as unsorted
        self._df_sorted[s_tuple] = False

    def get(self, key, key_and_value: bool=False):
        """
        Return elements of different kinds, matching the totally or partially specified key.

        :param key: A dictionary with all or part of the key of elements to be retrieved
        :param key_and_value: If True, return a list of tuple (key, value). If not, return a list of values
        :return: A list of elements which can be (key, value) or "value", depending on the parameter "key_and_value"
        """
        keys = [k for k in key]
        s_tuple = tuple(sorted(keys))
        if s_tuple not in self._dfs:
            # Try partial match
            s = set(keys)
            df = []
            for s2 in self._key_lst:
                if s.issubset(s2):
                    s_tuple = tuple(sorted(s2))
                    df.append((self._dfs[s_tuple], s_tuple))
        else:
            df = [(self._dfs[s_tuple], s_tuple)]

        if df:
            res = []
            for df_, s_tuple in df:
                try:
                    if case_sensitive:
                        df_key = tuple([key[k] if k in key else slice(None) for k in s_tuple])
                    else:
                        df_key = tuple([(key[k] if k.startswith("__") else key[k].lower()) if k in key else slice(None) for k in s_tuple])

                    if s_tuple not in self._df_sorted or not self._df_sorted[s_tuple]:
                        df_.sort_index(ascending=True, inplace=True)
                        self._df_sorted[s_tuple] = True

                    tmp = df_.loc[df_key, "value"]
                    if isinstance(tmp, pd.Series):
                        for i, v in enumerate(tmp):
                            if key_and_value:
                                k = {rk: rv for rk, rv in zip(s_tuple, tmp.index[i])}
                                res.append((k, v))
                            else:
                                res.append(v)
                    else:  # Single result, standardize to always (k, v)
                        if key_and_value:
                            k = {rk: rv for rk, rv in zip(s_tuple, df_key)}
                            res.append((k, tmp))
                        else:
                            res.append(tmp)
                except (IndexError, KeyError):
                    pass
                except Exception as e:
                    pass
            return res
        else:
            return []

    def delete(self, key):
        """
        Remove elements matching each of the keys, total or partial, passed to the method

        :param key:
        :return:
        """

        def delete_single(key_):
            """
            Remove elements matching the total or partial key

            :param key_:
            :return:
            """
            """
            Return elements of different kinds, matching the totally or partially specified key.
    
            :param key: A dictionary with all or part of the key of elements to be retrieved
            :param key_and_value: If True, return a list of tuple (key, value). If not, return a list of values
            :return: A list of elements which can be (key, value) or "value", depending on the parameter "key_and_value"
            """
            keys = [k_ for k_ in key_]
            s_tuple = tuple(sorted(keys))
            if s_tuple not in self._dfs:
                # Try partial match
                s = set(keys)
                df = []
                for s2 in self._key_lst:
                    if s.issubset(s2):
                        s_tuple = tuple(sorted(s2))
                        df.append((self._dfs[s_tuple], s_tuple))
            else:
                df = [(self._dfs[s_tuple], s_tuple)]

            if df:
                res = 0
                for df_, s_tuple in df:
                    if case_sensitive:
                        df_key = tuple([key_[k_] if k_ in key_ else slice(None) for k_ in s_tuple])
                    else:
                        df_key = tuple([(key_[k_] if k_.startswith("__") else key_[k_].lower()) if k_ in key_ else slice(None) for k_ in s_tuple])
                    try:
                        tmp = df_.loc[df_key, "value"]
                        if isinstance(tmp, pd.Series):
                            df_.drop(tmp.index, inplace=True)
                            res += len(tmp)
                        else:  # Single result, standardize to always (k, v)
                            df_.drop(df_key, inplace=True)
                            res += 1
                    except (IndexError, KeyError):
                        pass
                return res
            else:
                return 0

        if isinstance(key, list):
            res_ = 0
            for k in key:
                res_ += delete_single(k)
            return res_
        else:
            return delete_single(key)

    def to_pickable(self):
        # Convert to a jsonpickable structure
        out = {}
        for k in self._dfs:
            df = self._dfs[k]
            out[k] = df.to_dict()
        return out

    def from_pickable(self, inp):
        # Convert from the jsonpickable structure (obtained by "to_pickable") to the internal structure
        self._dfs = {}
        self._key_lst = []
        for t in inp:
            t_ = ast.literal_eval(t)
            self._key_lst.append(set(t_))
            inp[t]["value"] = {ast.literal_eval(k): v for k, v in inp[t]["value"].items()}
            df = pd.DataFrame(inp[t])
            df.index.names = t_
            self._dfs[t_] = df

        return self  # Allows the following: prd = PartialRetrievalDictionary().from_pickable(inp)


class PartialRetrievalDictionary:
    """
    A (k, v) storage where "k" is Dict of (str, str) and "v" is an object:

    reg = PartialRetrievalDictionary()
    reg.put(dict(a="1", b="3"), [4, 5, 9])

    The retrieval method, "get", is passed a partial key also in the form of a Dict of (str, str) and returns a list
    of objects matching the partial key:

    reg.get(dict(a="1")) -> [[4, 5, 9]]

    NOTE: when inserting, both "k" and "v" (object) should be unique

    """
    def __init__(self):
        # A dictionary of key-name to dictionaries, where the dictionaries are each of the values of the key and the
        # value is a set of IDs having that value
        # dict(key-name, dict(key-value, set(obj-IDs with that key-value))
        self._keys = {}
        # Dictionary from ID to the tuple (composite-key-elements dict, object)
        self._objs = {}
        self._rev_objs = {}  # From object to ID
        # Counter
        self._id_counter = 0

    def get(self, key, key_and_value=False, full_key=False, just_oid=False):
        """
        Retrieve one or more objects matching "key"
        If "key_and_value" is True, return not only the value, also matching key (useful for multiple matching keys)
        If "full_key" is True, zero or one objects should be the result
        :param key:
        :param full_key:
        :return: A list of matching elements
        """
        if True:
            # Lower case values
            # Keys can be all lower case, because they will be internal Key components, not specified by users
            if case_sensitive:
                key2 = {k.lower(): v for k, v in key.items()}
            else:
                key2 = {k.lower(): v if k.startswith("__") else v.lower() for k, v in key.items()}
        else:
            key2 = key

        sets = [self._keys.get(k, {}).get(v, set()) for k, v in key2.items()]

        # Find shorter set and Remove it from the list
        min_len = 1e30
        min_len_set_idx = None
        for i, s in enumerate(sets):
            if len(s) < min_len:
                min_len = len(s)
                min_len_set_idx = i
        min_len_set = sets[min_len_set_idx]
        del sets[min_len_set_idx]
        # Compute intersections
        result = min_len_set.intersection(*sets)
        if just_oid:
            return result

        # Obtain list of results
        if full_key and len(result) > 1:
            raise Exception("Zero or one results were expected. "+str(len(result)+" obtained."))
        if not key_and_value:
            return [self._objs[oid][1] for oid in result]
        else:
            return [self._objs[oid] for oid in result]

    def get_one(self, key, key_and_value=False, full_key=False, just_oid=False):
        results = self.get(key, key_and_value, full_key, just_oid)
        return results[0] if results else None

    def put(self, key: Dict[Tuple, object], value):
        """
        Insert implies the key does not exist
        Update implies the key exists
        Upsert does not care

        :param key:
        :param value:
        :return:
        """
        ptype = 'i'  # 'i', 'u', 'ups' (Insert, Update, Upsert)
        if True:
            # Lower case values
            # Keys can be all lower case, because they will be internal Key components, not specified by users
            if case_sensitive:
                key2 = {k.lower(): v for k, v in key.items()}
            else:
                key2 = {k.lower(): v if k.startswith("__") else v.lower() if isinstance(v, str) else v for k, v in key.items()}
        else:
            key2 = key
        # Arrays containing key: values "not-present" and "present"
        not_present = []  # List of tuples (dictionary of key-values, value to be stored)
        present = []  # List of sets storing IDs having same key-value
        for k, v in key2.items():
            d = self._keys.get(k, {})
            if len(d) == 0:
                self._keys[k] = d
            if v not in d:  # A new value "v" for subkey "k"
                not_present.append((d, v))
            else:
                present.append(d.get(v))

        if len(not_present) > 0:
            is_new = True
        else:
            if len(present) > 1:
                is_new = len(present[0].intersection(*present[1:])) == 0
            elif len(present) == 1:
                is_new = len(present) == 0
            else:
                is_new = False

        # Insert, Update or Upsert
        if is_new:  # It seems to be an insert
            # Check
            if ptype == 'u':
                raise Exception("Key does not exist")
            # Insert
            if value in self._rev_objs:
                oid = self._rev_objs[value]
            else:
                self._id_counter += 1
                oid = self._id_counter
                self._objs[oid] = (key, value)
                self._rev_objs[value] = oid

            # Insert

            # Add the subkey value to the subkey dictionaries where it is not present
            for d, v in not_present:
                s = set()
                d[v] = s
                s.add(oid)
            # Register the object into subkey-value dictionaries
            for s in present:
                s.add(oid)
        else:
            if ptype == 'i':
                raise Exception("Key '+"+str(key2)+"' already exists")
            # Update
            # Find the ID for the key
            res = self.get(key, just_oid=True)
            if len(res) != 1:
                raise Exception("Only one result expected")
            # Update value (key is the same, ID is the same)
            self._objs[res[0]] = value

    def delete(self, key):
        def delete_single(key):
            if True:
                # Lower case values
                # Keys can be all lower case, because they will be internal Key components, not specified by users
                if case_sensitive:
                    key2 = {k.lower(): v for k, v in key.items()}
                else:
                    key2 = {k.lower(): v if k.startswith("__") else v.lower() for k, v in key.items()}
            else:
                key2 = key

            # Get IDs
            oids = self.get(key, just_oid=True)
            if len(oids) > 0:
                # From key_i: value_i remove IDs (set difference)
                for k, v in key2.items():
                    d = self._keys.get(k, None)
                    if d:
                        s = d.get(v, None)
                        if s:
                            s2 = s.difference(oids)
                            d[v] = s2
                            if not s2:
                                del d[v]  # Remove the value for the key

                # Delete oids
                for oid in oids:
                    del self._objs[oid]

                return len(oids)
            else:
                return 0

        if isinstance(key, list):
            res_ = 0
            for k in key:
                res_ += delete_single(k)
            return res_
        else:
            return delete_single(key)

    def to_pickable(self):
        # Convert to a jsonpickable structure
        return dict(keys=self._keys, objs=self._objs, cont=self._id_counter)

    def from_pickable(self, inp):
        self._keys = inp["keys"]
        self._objs = {int(k): v for k, v in inp["objs"].items()}
        self._rev_objs = {v[1]: k for k, v in self._objs.items()}
        self._id_counter = inp["cont"]

        return self  # Allows the following: prd = PartialRetrievalDictionary().from_pickable(inp)


# #####################################################################################################################
# >>>> EXTERNAL DATASETS <<<<
# #####################################################################################################################


def get_statistical_dataset_structure(source, dset_name, local_datasets=None):
    from nexinfosys.ie_imports.data_sources.ad_hoc_dataset import AdHocDatasets
    # Register AdHocDatasets
    if local_datasets:
        # Register AdHocSource, which needs the current state
        adhoc = AdHocDatasets(local_datasets)
        nexinfosys.data_source_manager.register_datasource_manager(adhoc)

    # Obtain DATASET: Datasource -> Database -> DATASET -> Dimension(s) -> CodeList (no need for "Concept")
    dset = nexinfosys.data_source_manager.get_dataset_structure(source, dset_name, local_datasets)

    # Generate "dims", "attrs" and "meas" from "dset"
    dims = create_dictionary()  # Each dimension has a name, a description and a code list
    attrs = create_dictionary()
    meas = create_dictionary()
    for dim in dset.dimensions:
        if dim.is_measure:
            meas[dim.code] = None
        else:
            # Convert the code list to a dictionary
            if dim.get_hierarchy():
                cl = dim.get_hierarchy().to_dict()
            else:
                cl = None
            dims[dim.code] = SDMXConcept("dimension", dim.code, dim.is_time, "", cl)

    time_dim = False

    lst_dim = []

    for l in dims:
        lst_dim_codes = []
        if dims[l].istime:
            time_dim = True
        else:
            lst_dim.append((dims[l].name, lst_dim_codes))

        if dims[l].code_list:
            for c, description in dims[l].code_list.items():
                lst_dim_codes.append((c, description))

    if time_dim:
        lst_dim.append(("StartPeriod", None))
        lst_dim.append(("EndPeriod", None))

    # Unregister AdHocDatasets
    if local_datasets:
        nexinfosys.data_source_manager.unregister_datasource_manager(adhoc)

    return lst_dim, (dims, attrs, meas)


def check_dataset_exists(dset_name, local_datasets=None):
    from nexinfosys.ie_imports.data_source_manager import DataSourceManager
    if len(dset_name.split(".")) == 2:
        source, d_set = dset_name.split(".")
    else:
        d_set = dset_name
    res = DataSourceManager.obtain_dataset_source(d_set, local_datasets)
    return res is not None


def obtain_dataset_metadata(dset_name, source=None, local_datasets=None):
    from nexinfosys.ie_imports.data_source_manager import DataSourceManager
    d_set = dset_name
    if not source:
        if len(dset_name.split(".")) == 2:
            source, d_set = dset_name.split(".")
        else:
            source = DataSourceManager.obtain_dataset_source(d_set, local_datasets)

    _, metadata = get_statistical_dataset_structure(source, d_set, local_datasets)

    return metadata

# #####################################################################################################################
# >>>> DECORATORS <<<<
# #####################################################################################################################


class Memoize:
    """
    Cache of function calls (non-persistent, non-refreshable)
    """
    def __init__(self, fn):
        self.fn = fn
        self.memo = {}

    def __call__(self, *args):
        if args not in self.memo:
            self.memo[args] = self.fn(*args)
        return self.memo[args]


class Memoize2(object):
    """cache the return value of a method

    This class is meant to be used as a decorator of methods. The return value
    from a given method invocation will be cached on the instance whose method
    was invoked. All arguments passed to a method decorated with memoize must
    be hashable.

    If a memoized method is invoked directly on its class the result will not
    be cached. Instead the method will be invoked like a static method:
    class Obj(object):
        @memoize
        def add_to(self, arg):
            return self + arg
    Obj.add_to(1) # not enough arguments
    Obj.add_to(1, 2) # returns 3, result is not cached
    """

    def __init__(self, func):
        self.func = func

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self.func
        return partial(self, obj)

    def __call__(self, *args, **kw):
        obj = args[0]
        try:
            cache = obj.__cache
        except AttributeError:
            cache = obj.__cache = {}
        key = (self.func, args[1:], frozenset(kw.items()))
        try:
            res = cache[key]
        except KeyError:
            res = cache[key] = self.func(*args, **kw)
        return res


# #####################################################################################################################
# >>>> MAPPING <<<<
# #####################################################################################################################


# Cython version is in module "helper_accel.pyx"
def augment_dataframe_with_mapped_columns(
        df: DataFrame,
        dict_of_maps: Dict[str, Tuple[str, List[Dict]]],
        measure_columns: List[str]) -> DataFrame:
    """
    Elaborate a pd.DataFrame from the input DataFrame "df" and
    "dict_of_maps" which is a dictionary of "source_column" to a tuple ("destination_column", map)
    where map is of the form:
        [ {origin category: [{d: destination category, w: weight assigned to destination category}] } ]

    Support not only "Many to One" (ManyToOne) but also "Many to Many" (ManyToMany)

    :param df: pd.DataFrame to process
    :param dict_of_maps: dictionary from "source" to a tuple of ("destination", "map"), see previous introduction
    :param measure_columns: list of measure column names in "df"
    :return: The pd.DataFrame resulting from the mapping
    """
    mapped_cols = {}  # A dict from mapped column names to column index in "m"
    measure_cols = {}  # A dict from measure columns to column index in "m". These will be the columns affected by the mapping weights
    non_mapped_cols = {}  # A dict of non-mapped columns (the rest)
    for i, c in enumerate(df.columns):
        if c in dict_of_maps:
            mapped_cols[c] = i
        elif c in measure_columns:
            measure_cols[c] = i
        else:
            non_mapped_cols[c] = i

    # "np.ndarray" from "pd.DataFrame" (no index, no column labels, only values)
    m = df.values

    # First pass is just to obtain the size of the target ndarray
    ncols = 2*len(mapped_cols) + len(non_mapped_cols) + len(measure_cols)
    nrows = 0
    for r in range(m.shape[0]):
        # Obtain each code combination
        n_subrows = 1
        for c_name, c in mapped_cols.items():
            map_ = dict_of_maps[c_name][1]
            code = m[r, c]
            n_subrows *= len(map_[code])
        nrows += n_subrows

    # Second pass, to elaborate the elements of the destination array
    new_cols_base = len(mapped_cols)
    non_mapped_cols_base = 2*len(mapped_cols)
    measure_cols_base = non_mapped_cols_base + len(non_mapped_cols)

    # Output matrix column names
    col_names = [col for col in mapped_cols]
    col_names.extend([dict_of_maps[col][0] for col in mapped_cols])
    col_names.extend([col for col in non_mapped_cols])
    col_names.extend([col for col in measure_cols])
    assert len(col_names) == ncols

    # Output matrix
    mm = np.empty((nrows, ncols), dtype=object)

    # For each ROW of ORIGIN matrix
    row = 0  # Current row in output matrix
    for r in range(m.shape[0]):
        # Obtain combinations from current codes
        lst = []
        n_subrows = 1
        for c_name, c in mapped_cols.items():
            map_ = dict_of_maps[c_name][1]
            code = m[r, c]
            n_subrows *= len(map_[code])
            lst.append(map_[code])

        combinations = list(itertools.product(*lst))

        for icomb in range(n_subrows):
            combination = combinations[icomb]
            # Mapped columns
            # At the same time, compute the weight for the measures
            w = 1.0
            for i, col in enumerate(mapped_cols.keys()):
                mm[row+icomb, i] = m[r, mapped_cols[col]]
                mm[row+icomb, new_cols_base+i] = ifnull(combination[i]["d"], '')
                if combination[i]["w"]:
                    w *= float(combination[i]["w"])

            # Fill the measures
            for i, col in enumerate(measure_cols.keys()):
                mm[row+icomb, measure_cols_base+i] = w * m[r, measure_cols[col]]

            # Non-mapped columns
            for i, col in enumerate(non_mapped_cols.keys()):
                mm[row+icomb, non_mapped_cols_base+i] = m[r, non_mapped_cols[col]]

        row += n_subrows

    # Now elaborate a DataFrame back
    tmp = pd.DataFrame(data=mm, columns=col_names)
    for c in measure_columns:
        tmp[c] = tmp[c].astype(np.float)

    return tmp


def is_boolean(v):
    return v.lower() in ["true", "false"]


def to_boolean(v):
    return v.lower() == "true"


def is_integer(v):
    try:
        int(v)
        return True
    except ValueError:
        return False


def to_integer(v):
    return int(v)


def is_float(v):
    try:
        float(v)
        return True
    except ValueError:
        return False


def to_float(v):
    return float(v)


def is_datetime(v):
    try:
        from dateutil.parser import parse
        parse(v)
        return True
    except ValueError:
        return False


def to_datetime(v):
    from dateutil.parser import parse
    return parse(v)


def is_url(v):
    """
    From https://stackoverflow.com/a/36283503
    """
    min_attributes = ('scheme', 'netloc')
    qualifying = min_attributes
    token = urllib.parse.urlparse(v)
    return all([getattr(token, qualifying_attr) for qualifying_attr in qualifying])


def to_url(v):
    return urllib.parse.urlparse(v)


def is_uuid(v):
    try:
        UUID(v, version=4)
        return True
    except ValueError:
        return False


def to_uuid(v):
    return UUID(v, version=4)


def is_category(v):
    # TODO Get all hierarchies, get all categories from all hierarchies, find if "v" is one of them
    return False


def to_category(v):
    # TODO
    return None #  Taxon  # Return some Taxon


def is_geo(v):
    # TODO Check if "v" is a GeoJSON, or a reference to a GeoJSON
    return None


def to_geo(v):
    return None


def is_str(v):
    return True


def to_str(v):
    return str(v)

# #####################################################################################################################
# >>>> LOAD DATASET FROM URL INTO PD.DATAFRAME <<<<
# #####################################################################################################################


def wv_create_client_and_resource_name(location, wv_user=None, wv_password=None, wv_host_name=None):
    import webdav.client as wc
    if not wv_host_name:
        wv_host_name = get_global_configuration_variable("FS_SERVER") \
            if get_global_configuration_variable("FS_SERVER") else "nextcloud.data.magic-nexus.eu"

    parts = location.split("/")
    for i, p in enumerate(parts):
        if p == wv_host_name:
            url = "/".join(parts[:i + 1])
            fname = "/" + "/".join(parts[i + 1:])
            break

    options = {
        "webdav_hostname": url,
        "webdav_login": wv_user if wv_user else get_global_configuration_variable("FS_USER"),
        "webdav_password": wv_password if wv_password else get_global_configuration_variable("FS_PASSWORD")
    }
    client = wc.Client(options)
    return client, fname


def wv_check_path(location, wv_user=None, wv_password=None, wv_host_name=None):
    """
    Check if the location exists, and if it is a directory or a file

    :param location:
    :param user:
    :param password:
    :return: 0 -> does not exist; 1 -> directory; 2 -> file
    """
    client, fname = wv_create_client_and_resource_name(location, wv_user, wv_password, wv_host_name)
    if client.check(fname):
        retval = 1 if client.is_dir(fname) else 2
    else:
        retval = 0
    client.free()
    return retval


def wv_create_directory(location, wv_user=None, wv_password=None, wv_host_name=None):
    """
    Create directory (if it exists do nothing)

    :param location:
    :param wv_user:
    :param wv_password:
    :param wv_host_name:
    :return:

    """
    client, fname = wv_create_client_and_resource_name(location, wv_user, wv_password, wv_host_name)
    client.mkdir(fname)
    client.free()


def wv_upload_file(data: io.BytesIO, location: str, wv_user=None, wv_password=None, wv_host_name=None):
    """
    Uploads a bytes array into a location on a WebDAV server
    Directory must exist
    :param data: The data to upload
    :param location:
    :param wv_user:
    :param wv_password:
    :param wv_host_name:
    :return:
    """
    client, fname = wv_create_client_and_resource_name(location, wv_user, wv_password, wv_host_name)
    client.upload_from(data, fname)
    client.free()


def wv_download_file(location, wv_user=None, wv_password=None, wv_host_name=None):
    """
    WebDAV download a file

    :param location:
    :param wv_user:
    :param wv_password:
    :param wv_host_name:
    :return:
    """
    client, fname = wv_create_client_and_resource_name(location, wv_user, wv_password, wv_host_name)
    with tempfile.NamedTemporaryFile(delete=True) as temp:
        client.download_sync(remote_path=fname, local_path=temp.name)
        f = open(temp.name, "rb")
        data = io.BytesIO(f.read())
        f.close()
    client.free()
    return data


def download_with_pycurl(location):
    headers = {}

    def header_function(header_line):
        # HTTP standard specifies that headers are encoded in iso-8859-1.
        # On Python 2, decoding step can be skipped.
        # On Python 3, decoding step is required.
        header_line = header_line.decode('iso-8859-1')

        # Header lines include the first status line (HTTP/1.x ...).
        # We are going to ignore all lines that don't have a colon in them.
        # This will botch headers that are split on multiple lines...
        if ':' not in header_line:
            return

        # Break the header line into header name and value.
        name, value = header_line.split(':', 1)

        # Remove whitespace that may be present.
        # Header lines include the trailing newline, and there may be whitespace
        # around the colon.
        name = name.strip()
        value = value.strip()

        # Header names are case insensitive.
        # Lowercase name here.
        name = name.lower()

        # Now we can actually record the header name and value.
        # Note: this only works when headers are not duplicated, see below.
        headers[name] = value

    data = io.BytesIO()
    c = pycurl.Curl()
    c.setopt(c.URL, location)
    c.setopt(c.FOLLOWLOCATION, True)
    c.setopt(c.HEADERFUNCTION, header_function)
    c.setopt(c.WRITEDATA, data)
    c.perform()
    status = c.getinfo(c.RESPONSE_CODE)
    c.close()

    return status, headers, data


def download_file(location, wv_user=None, wv_password=None, wv_host_name=None):
    """
    Download a file from the specified URL location.

    It recognizes MAGIC's Nextcloud (WebDAV) instance AND Google Drive URL's
    Of course, it should work with Zenodo.
    Google Drive URL's are assumed to be Spreadsheets (both XLSX and Google Calc are considered)

    :param location:
    :param wv_host_name: WebDav host name part. Example: "nextcloud.data.magic-nexus.eu"
    :return: A BytesIO object with the contents of the file
    """
    pr = urlparse(location)
    if pr.scheme != "":
        # Load from remote site
        fragment = ""
        if "#" in location:
            pos = location.find("#")
            fragment = location[pos + 1:]
            location = location[:pos]
        if not wv_host_name:
            wv_host_name = get_global_configuration_variable("FS_SERVER")\
                if get_global_configuration_variable("FS_SERVER") else "nextcloud.data.magic-nexus.eu"
        if pr.netloc.lower() == wv_host_name:
            data = wv_download_file(location, wv_user, wv_password, wv_host_name)
        elif pr.netloc.lower() == "docs.google.com" or pr.netloc.lower() == "drive.google.com":
            # TODO From mail: Aug 11, 2021, 6:47 AM
            #   with subject: [Action Required] Drive API requires updates to your code before Sep 13, 2021
            #   13 of September
            #      https://developers.google.com/drive/api/v3/reference/files
            #      https://developers.google.com/drive/api/v3/resource-keys
            #

            # Google Drive. CSV and XLSX files supported (if Google Sheets, an Export to XLSX is done)
            # Extract file id from the URL
            import re
            m = re.match(r".*[^-\w]([-\w]{33,})[^-\w]?.*", location)
            file_id = m.groups()[0]
            if "com/file" not in location:
                url = f"https://docs.google.com/spreadsheets/d/{file_id}/export?format=xlsx"  # &id={file_id}"
            else:
                url = f"https://docs.google.com/uc?export=download&id={file_id}"
            # resp = requests.get(url, headers={'Cache-Control': 'no-cache', 'Pragma': 'no-cache'}, allow_redirects=True)  # headers={'Cache-Control': 'no-cache', 'Pragma': 'no-cache'}
            status_code, headers, data = download_with_pycurl(url)
            logging.debug(f'curl -L "{url}" >> out.xlsx')
            if status_code == 200 and "text/html" not in headers["content-type"]:
            # if resp.status_code == 200 and "text/html" not in resp.headers["Content-Type"]:
            #     data = io.BytesIO(resp.content)
                pass
            else:
                credentials_file = get_global_configuration_variable("GAPI_CREDENTIALS_FILE")
                token_file = get_global_configuration_variable("GAPI_TOKEN_FILE")
                data = download_xlsx_file_id(credentials_file, token_file, file_id)
            # with open("/home/rnebot/Downloads/out2.xlsx", "wb") as nf:
            #     nf.write(data.getvalue())
        else:
            data = urllib.request.urlopen(location).read()
            data = io.BytesIO(data)
    else:
        data = urllib.request.urlopen(location).read()
        data = io.BytesIO(data)

    return data


def load_dataset(location: str=None):
    """
    Loads a dataset into a DataFrame
    If the dataset is present, it decompresses it in memory to obtain one of the four datasets per file
    If the dataset is not downloaded, downloads it and decompresses into the corresponding version directory
    :param location: URL of the dataset data
    :return: pd.DataFrame
    """

    if not location:
        df = None
    else:
        data = download_file(location)
        fragment = ""
        if "#" in location:
            pos = location.find("#")
            fragment = location[pos + 1:]
            location = location[:pos-1]
        # Then, try to read it
        t = mimetypes.guess_type(location, strict=True)
        if t[0] == "text/csv" or "com/file" in location:
            data.seek(0)
            df = pd.read_csv(data, comment="#")
        elif t[0] == "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet" or t[0] == "application/vnd.ms-excel":
            if fragment:
                df = pd.read_excel(data, sheet_name=fragment)
            else:
                df = pd.read_excel(data)

    return df


def any_error_issue(issues):
    """
    Just iterate through a list of Issues (of the three types) to check if there is any error

    :param issues:
    :return:
    """
    any_error = False
    for i in issues:
        from nexinfosys.command_generators import Issue, IType
        if isinstance(i, dict):
            if i["type"] == 3:
                any_error = True
        elif isinstance(i, tuple):
            if i[0] == 3:  # Error
                any_error = True
        elif isinstance(i, Issue):
            if i.itype == IType.ERROR:  # Error
                any_error = True
    return any_error


def prepare_dataframe_after_external_read(ds, df, cmd_name):
    from nexinfosys.command_generators import IType, Issue, IssueLocation  # Declared here to avoid circular "import"
    issues = []
    dims = set()  # Set of dimensions, to index Dataframe on them
    cols = []  # Same columns, with exact case (matching Dataset object)
    orig_cols = []  # Original column names appearing in the target Dataset
    for c in df.columns:
        for d in ds.dimensions:
            if strcmp(c, d.code):
                orig_cols.append(c)
                cols.append(d.code)  # Exact case
                if not d.is_measure:
                    dims.add(d.code)
                    num_column = df.dtypes[c] in [np.int64, np.float]
                    # NaN values in dimensions replaced by ""
                    df[c] = df[c].fillna('')
                    # Convert to string if it is numeric
                    if num_column:
                        df[c] = df[c].astype(str)
                break
        else:
            del df[c]
            issue = Issue(itype=IType.WARNING,
                          description=f"Column '{c}' not found in the definition of Dataset '{ds.code}'. Skipping it.",
                          location=IssueLocation(sheet_name=cmd_name,
                                                 row=None, column=None))
            issues.append(issue)
    df = df[orig_cols]  # Remove unused columns
    df.columns = cols  # Rename columns
    df.set_index(list(dims), inplace=True)  # Index by Dimension Concepts
    # Check Index: unique
    tmp = np.where(df.index.duplicated(keep=False))
    if len(tmp[0]) > 0:
        for i in np.nditer(tmp):
            issue = Issue(itype=IType.ERROR,
                          description=f"Duplicated row '{i+2}' in the data of dataset '{ds.code}' with value: {df.iloc[i]}",
                          location=IssueLocation(sheet_name=cmd_name,
                                                 row=i+2, column=None))
            issues.append(issue)
    df.reset_index(inplace=True)  # If not, information stored in the index is lost (maybe after serialization/deserialization)

    return issues


# #####################################################################################################################
# >>>> DATAFRAME <<<<
# #####################################################################################################################

def get_dataframe_copy_with_lowercase_multiindex(dataframe: DataFrame) -> DataFrame:
    """
    Create a copy of an input MultiIndex dataframe where all the index values have been lowercased.
    :param dataframe: a MultiIndex dataframe
    :return: A copy of the input dataframe with lowercased index values.
    """
    # df = dataframe.copy()
    levels = [dataframe.index.get_level_values(n).astype(str).str.lower() for n in range(dataframe.index.nlevels)]
    df = pd.DataFrame(index=pd.MultiIndex.from_arrays(levels))
    # df.index = pd.MultiIndex.from_arrays(levels)
    return df

# #####################################################################################################################
# >>>> OTHER STUFF <<<<
# #####################################################################################################################


def str2bool(v: str):
    return str(v).lower() in ("yes", "true", "t", "1")


def ascii2uuid(s: str) -> str:
    """
    Convert an ASCII string to an UUID hexadecimal string
    :param s: an ASCII string
    :return: an UUID hexadecimal string
    """
    return str(uuid.UUID(bytes=base64.a85decode(s)))


T = TypeVar('T')
S = TypeVar('S')


def ifnull(var: Optional[T], val: Optional[S]) -> Union[T, Optional[S]]:
    """ Returns first value if not None, otherwise returns second value """
    if var is None:
        return val
    return var


def head(in_list: List[T]) -> Optional[T]:
    """
    Returns the head element of the list or None if the list is empty.
    :param in_list: The input list
    :return: The head element of the list or None
    """
    if in_list:
        return in_list[0]
    else:
        return None


def first(iterable: Iterable[T],
          condition: Callable[[T], bool] = lambda x: True,
          default: Optional[T] = None) -> Optional[T]:
    """
    Returns the first item in the `iterable` that satisfies the `condition`.
    If the condition is not given, returns the first item of the iterable.

    Returns the `default` value if no item satisfying the condition is found.

    >>> first( (1,2,3), condition=lambda x: x % 2 == 0)
    2
    >>> first(range(3, 100))
    3
    >>> first( () )
    None
    >>> first( (), default="Some" )
    Some
    """
    return next((x for x in iterable if condition(x)), default)


def translate_case(current_names: List[str], new_names: List[str]) -> List[str]:
    """
    Translate the names in the current_names list according the existing names in the new_names list that
    can have a different case.
    :param current_names: a list of names to translate
    :param new_names: a list of new names to use
    :return: the current_names list where some or all of the names are translated
    """
    new_names_dict = {name.lower(): name for name in new_names}
    translated_names = [new_names_dict.get(name.lower(), name) for name in current_names]
    return translated_names


def values_of_nested_dictionary(d: Dict)-> List:
    for v in d.values():
        if not isinstance(v, Dict):
            yield v
        else:
            yield from values_of_nested_dictionary(v)


def name_and_id_dict(obj: object) -> Optional[Dict]:
    if obj:
        return {"name": obj.name, "id": obj.uuid}
    else:
        return None


def get_value_or_list(current_value, additional_value):
    """
    Add a new value to another existing value/s. If a value doesn't exist it returns the new value otherwise
    returns a list with the existing value/s and the new value.
    :param current_value: the current value
    :param additional_value: the new value
    :return: a single value or a list
    """
    if current_value:
        if isinstance(current_value, list):
            return current_value + [additional_value]
        else:
            return [current_value, additional_value]
    else:
        return additional_value


def class_full_name(c: Type) -> str:
    """ Get the full name of a class """
    module = c.__module__
    if module is None:
        return c.__name__
    else:
        return module + '.' + c.__name__


def object_full_name(o: object) -> str:
    """ Get the full class name of an object """
    return class_full_name(o.__class__)


def split_and_strip(s: str, sep=",") -> List[str]:
    """Split a string representing a comma separated list of strings into a list of strings
    where each element has been stripped. If the string has no elements an empty list is returned."""
    string_list: List[str] = []
    if s is not None and isinstance(s, str):
        string_list = [s.strip() for s in s.split(sep)]

    return string_list


def precedes_in_list(lst: List[T], elem1: Optional[T], elem2: Optional[T]) -> bool:
    """ Check if an element comes before another inside a list """
    return elem1 in lst and (elem2 not in lst or lst.index(elem1) < lst.index(elem2))


def replace_string_from_dictionary(s: str, d: Dict[str, str]) -> str:
    """
    Replace in a string all the words defined as keys in the dictionary with the associated value.
    Note: values of the dictionary are sorted by length before replacement in order to avoid replacing on top of a replaced word.

    :param s: input string to transform with replacements
    :param d: dictionary with old_word -> new_word scheme
    :return: a string with replacements applied
    """
    for k, v in sorted(d.items(), key=lambda item: len(item[1])):
        logging.debug(v)
        s = re.sub(r"\b" + k + r"\b", v, s)
    return s

# #####################################################################################################################
# >>>> CUSTOM DATA TYPES <<<<
# #####################################################################################################################

FloatOrStringT = Union[str, float]


class FloatOrString:
    @staticmethod
    def to_float(value: Optional[FloatOrStringT]) -> Optional[FloatOrStringT]:
        if value is None:
            return None
        try:
            return float(value)
        except ValueError:
            return value

    @staticmethod
    def multiply(a: Optional[FloatOrStringT], b: Optional[FloatOrStringT]) -> Optional[FloatOrStringT]:
        value_a = FloatOrString.to_float(a)
        value_b = FloatOrString.to_float(b)

        if value_a is None:
            return value_b

        if value_b is None:
            return value_a

        if isinstance(value_a, float) and isinstance(value_b, float):
            return value_a * value_b
        else:
            return f"({value_a})*({value_b})"

    @staticmethod
    def multiply_with_float(a: FloatOrStringT, b: float) -> FloatOrStringT:
        value_a = FloatOrString.to_float(a)

        if isinstance(value_a, float):
            return value_a * b
        else:
            return f"({value_a})*{b}"


class UnitConversion:
    @staticmethod
    def convert(value: FloatOrStringT, from_unit: str, to_unit: str) -> FloatOrStringT:
        ratio = UnitConversion.ratio(from_unit, to_unit)
        return FloatOrString.multiply_with_float(value, ratio)

    @staticmethod
    def ratio(from_unit: str, to_unit: str) -> float:
        return nexinfosys.ureg(from_unit).to(nexinfosys.ureg(to_unit)).magnitude

    @staticmethod
    def get_scaled_weight(weight: FloatOrStringT,
                          source_from_unit: str, source_to_unit: Optional[str],
                          target_from_unit: Optional[str], target_to_unit: str) -> FloatOrStringT:
        ratio = 1.0
        if source_to_unit:
            ratio *= UnitConversion.ratio(source_from_unit, source_to_unit)

        if target_from_unit:
            ratio *= UnitConversion.ratio(target_from_unit, target_to_unit)

        return FloatOrString.multiply_with_float(weight, ratio)


class ArithmeticOperator(Enum):
    ADD = ("+", add)
    MUL = ("*", mul)
    SUB = ("-", sub)
    DIV = ("/", truediv)
    ABS = ("abs", abs)

    def __init__(self, symbol: str, operation: Callable):
        self.symbol = symbol
        self.operation = operation


class ArithmeticOperand:
    def __init__(self, name: str, value: float):
        self.name = name
        self.value = value

    def __str__(self):
        return self.name


class ArithmeticExpression:
    def __init__(self, operator: ArithmeticOperator, operands: List[Union[ArithmeticOperand, 'ArithmeticExpression']], value: float):
        self.operator = operator
        self.operands = operands
        self.value = value

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        if len(self.operands) == 1:
            return f"{self.operator.symbol}({self.operands[0]})"
        else:
            return self.operator.symbol.join([f"({o})" for o in self.operands])


def brackets(exp: str) -> str:
    """ Surround a string expression with brackets """
    return "(" + exp + ")"


class FloatExp(SupportsFloat):
    """ Wrapper of the Float data type which includes a name and a string expression that will grow
        when operating with other objects """

    ValueWeightPair = Tuple[Optional['FloatExp'], Optional['FloatExp']]

    def __init__(self, val: Union[float, int], name: Optional[str] = None,
                 exp: Union[str, ArithmeticExpression] = None):
        assert (isinstance(val, float) or isinstance(val, int))
        assert (name is None or isinstance(name, str))
        assert (exp is None or isinstance(exp, str) or isinstance(exp, ArithmeticExpression))

        self.val = float(val)

        if exp is None:
            self.exp = ArithmeticOperand(str(self.val), self.val)
        elif isinstance(exp, str):
            self.exp = ArithmeticOperand(exp, self.val)
        else:
            self.exp = exp

        self.name = str(self.exp) if name is None else name

    def __float__(self) -> float:
        """ Needed for abc SupportsFloat used in match.isclose() """
        return self.val

    def __round__(self, n=None):
        exp = f"round({self.name}{'' if n is None else ',' + str(n)})"
        return FloatExp(float(round(self.val, n)), exp, self.exp)

    def __abs__(self):
        return self._operate_unary(ArithmeticOperator.ABS)

    @staticmethod
    def get_float(f: Union[float, 'FloatExp']) -> float:
        return f.val if isinstance(f, FloatExp) else f

    @staticmethod
    def compute_weighted_addition(addends: List[ValueWeightPair]) -> Optional['FloatExp']:
        filtered_addends = [(v, w) for v, w in addends if v]

        if len(filtered_addends) == 1:
            value, weight = filtered_addends[0]
            if weight and weight != 1.0:
                return value * weight
            else:
                return FloatExp(value.val, value.name, value.name)
        elif len(filtered_addends) > 1:
            values = [value * weight
                      if weight and weight != 1.0
                      else FloatExp(value.val, value.name, value.name)
                      for value, weight in filtered_addends]
            value = reduce(add, [value.val for value in values])
            exp = ArithmeticExpression(ArithmeticOperator.ADD, [v.exp for v in values], value)
            return FloatExp(value, str(exp), exp)
        else:
            return None

    def _operate_unary(self, operator: ArithmeticOperator) -> 'FloatExp':
        value = operator.operation(self.val)
        new_exp = ArithmeticExpression(operator, [ArithmeticOperand(self.name, self.val)], value)
        return FloatExp(value, str(new_exp), new_exp)

    def _operate_binary(self, operator: ArithmeticOperator, other: 'FloatExp') -> 'FloatExp':
        value = operator.operation(self.val, other.val)
        new_exp = ArithmeticExpression(operator, [ArithmeticOperand(self.name, self.val),
                                                  ArithmeticOperand(other.name, other.val)], value)
        return FloatExp(value, str(new_exp), new_exp)

    def __add__(self, other: 'FloatExp') -> 'FloatExp':
        return self._operate_binary(ArithmeticOperator.ADD, other)

    def __sub__(self, other: 'FloatExp') -> 'FloatExp':
        return self._operate_binary(ArithmeticOperator.SUB, other)

    def __mul__(self, other: 'FloatExp') -> 'FloatExp':
        return self._operate_binary(ArithmeticOperator.MUL, other)

    def __truediv__(self, other: 'FloatExp') -> 'FloatExp':
        return self._operate_binary(ArithmeticOperator.DIV, other)

    def __eq__(self, other: Union[float, 'FloatExp']) -> bool:
        return self.val == FloatExp.get_float(other)

    def __gt__(self, other: Union[float, 'FloatExp']) -> bool:
        return self.val > FloatExp.get_float(other)

    def __str__(self) -> str:
        return f'Value = {self.val}, Name = {self.name}, Expression = "{self.exp}"'

    def __repr__(self):
        return str(self)


def get_interfaces_and_weights_from_expression(exp: Union[ArithmeticOperand, ArithmeticExpression]) -> List[Tuple[str, float]]:
    def is_interface(s: str) -> bool:
        return s.find(":") > 0

    def find_interface(exp: Union[ArithmeticOperand, ArithmeticExpression]) -> Optional[Tuple[str, float]]:
        if isinstance(exp, ArithmeticOperand):
            if is_interface(exp.name):
                return exp.name, 1.0
            else:
                logging.debug(f"Operand not interface: {exp.name}")
        elif exp.operator == ArithmeticOperator.MUL:
            interfaces_list: List[Tuple[str, float]] = []
            weight: float = 1.0
            for o in exp.operands:
                i = find_interface(o)
                if i:
                    interfaces_list.append(i)
                else:
                    weight *= o.value

            if len(interfaces_list) == 1:
                return interfaces_list[0][0], interfaces_list[0][1] * weight
            elif len(interfaces_list) > 1:
                logging.debug(f"Error multiple interfaces: {interfaces_list}")
        else:
            logging.debug(f"Error operator not allowed: {exp.operator}")

        return None

    interfaces: List[Tuple[str, float]] = []
    if isinstance(exp, ArithmeticExpression) and exp.operator == ArithmeticOperator.ADD:
        for operand in exp.operands:
            i = find_interface(operand)
            if i:
                interfaces.append(i)
    else:
        i = find_interface(exp)
        if i:
            interfaces.append(i)

    return interfaces


def add_label_columns_to_dataframe(ds_name, df, prd):
    """
    Add columns containing labels describing codes in the input Dataframe
    The labels must be in the CodeHierarchies or CodeLists

    :param ds_name: Dataset name
    :param df: pd.Dataframe to enhance
    :param prd: PartialRetrievalDictionary
    :return: Enhanced pd.Dataframe
    """
    from nexinfosys.models.musiasem_concepts import Hierarchy
    # Merge with Taxonomy LABELS, IF available
    for col in df.columns:
        hs = prd.get(Hierarchy.partial_key(ds_name + "_" + col))
        hs2 = prd.get(Hierarchy.partial_key(col))
        if len(hs) == 1 or len(hs2) == 1:
            if len(hs) == 1:
                h = hs[0]
            else:
                h = hs2[0]
            nodes = h.get_all_nodes()
            tmp = []
            for nn in nodes:
                t = nodes[nn]
                tmp.append([t[0].lower(), t[1]])  # CSens
            if not nexinfosys.case_sensitive and df[col].dtype == 'O':
                df[col + "_l"] = df[col].str.lower()
                col = col + "_l"

                # Dataframe of codes and descriptions
                df_dst = pd.DataFrame(tmp, columns=['sou_rce', col + "_desc"])
                df = pd.merge(df, df_dst, how='left', left_on=col, right_on='sou_rce')
                del df['sou_rce']
                if not nexinfosys.case_sensitive:
                    del df[col]

    return df


def change_tuple_value(t: Tuple, index: int, value: Any) -> Tuple:
    lst = list(t)
    lst[index] = value
    return tuple(lst)


# if __name__ == '__main__':
#     import random
#     import string
#     from timeit import default_timer as timer
#
#     class Dummy:
#         def __init__(self, a):
#             self._a = a
#
#     def rndstr(n):
#         return random.choices(string.ascii_uppercase + string.digits, k=n)
#
#     prd = PartialRetrievalDictionary2()
#     ktypes = [("a", "b", "c"), ("a", "b"), ("a", "d"), ("a", "f", "g")]
#     # Generate a set of keys and empty objects
#     vals = []
#     print("Generating sample")
#     for i in range(30000):
#         # Choose random key
#         ktype = ktypes[random.randrange(len(ktypes))]
#         # Generate the element
#         vals.append(({k: ''.join(rndstr(6)) for k in ktype}, Dummy(rndstr(12))))
#
#     print("Insertion started")
#     df = pd.DataFrame()
#     start = timer()
#     # Insert each element
#     for v in vals:
#         prd.put(v[0], v[1])
#     stop = timer()
#     print(stop-start)
#
#     print("Reading started")
#
#     # Select all elements
#     start = timer()
#     # Insert each element
#     for v in vals:
#         r = prd.get(v[0], False)
#         if len(r) == 0:
#             raise Exception("Unexpected!")
#     stop = timer()
#     print(stop-start)
#
#     print("Deleting started")
#
#     # Select all elements
#     start = timer()
#     # Insert each element
#     for v in vals:
#         r = prd.delete(v[0])
#         if r == 0:
#             raise Exception("Unexpected!")
#     stop = timer()
#     print(stop-start)
#
#     print("Finished!!")


if __name__ == '__main__':
    # f = open("/home/rnebot/GoogleDrive/AA_MAGIC/FAOSTAT_analysis.xlsx", "rb")
    # data = io.BytesIO(f.read())
    # f.close()
    # wv_upload_file(data, "https://nextcloud.data.magic-nexus.eu/remote.php/webdav/NIS_beta/CS_format_examples/FAOSTAT_analysis.xlsx", "NIS_agent", "NIS_agent@1", "nextcloud.data.magic-nexus.eu")
    ret = download_file("https://sandbox.zenodo.org/record/536704/files/url_demo_2.zip?download=1#msm/geolayer.xlsx")
    # ret = download_file("https://nextcloud.data.magic-nexus.eu/remote.php/webdav/NIS_beta/CS_format_examples/08_caso_energia_eu_new_commands.xlsx", "NIS_agent", "NIS_agent@1")
    print(f"Longit: {len(ret.getvalue())}")




"""
Logic to manage the cache for external datasets

Features per dataset or per dataset source (shared by different datasets)
* Refresh times for the dataset
  * Conditions expiry time of cached information
* Dataset identification
* Eager (automatic) refresh or lazy refresh (on demand)
* Maximum size
"""


import json
from dotted.collection import DottedDict

from nexinfosys.command_generators import Issue, IssueLocation, IType
from nexinfosys.model_services import IExecutableCommand, get_case_study_registry_objects
from nexinfosys.common.helper import obtain_dataset_metadata, create_dictionary
from nexinfosys.models.musiasem_concepts import Mapping


# Combinatory of lists (for many-to-many mappings)
#
# import itertools
# a = [[(0.5, "a"),(4, "b"),(7, "c")],[(1, "b")],[(0.2, "z"),(0.1, "y"),(0.7, "x")]]
# list(itertools.product(*a))
#

def fill_map_with_all_origin_categories(dim, mapping):
    # Check all codes exist
    mapped_codes = set([d["o"] for d in mapping])
    all_codes = set([c for c in dim.code_list])
    for c in all_codes - mapped_codes:  # Loop over "unmapped" origin codes
        # This sentence MODIFIES map, so it is not necessary to return it
        mapping.append({"o": c, "to": [{"d": None, "w": 1.0}]})  # Map to placeholder, with weight 1

    return mapping


class HierarchyMappingCommand(IExecutableCommand):
    def __init__(self, name: str):
        self._name = name
        self._content = None

    def execute(self, state: "State"):
        def process_line(item):
            # Read variables
            mh_src_dataset = item.get("source_dataset", None)
            mh_src_hierarchy = item.get("source_hierarchy", None)
            mh_src_code = item.get("source_code", None)
            mh_dst_hierarchy = item.get("destination_hierarchy", None)
            mh_dst_code = item.get("destination_code", None)
            mh_weight = item.get("weight", 1.0)

            # Mapping name
            name = ((mh_src_dataset + ".") if mh_src_dataset else "") + mh_src_hierarchy + " -> " + mh_dst_hierarchy

            if name in mappings:
                issues.append(Issue(itype=IType.ERROR,
                                    description="The mapping '"+name+"' has been declared previously. Skipped.",
                                    location=IssueLocation(sheet_name=name, row=r, column=None)))
                return

            if name in local_mappings:
                d = local_mappings[name]
            else:
                d = DottedDict()
                local_mappings[name] = d
                d.name = name
                d.origin_dataset = mh_src_dataset
                d.origin_hierarchy = mh_src_hierarchy
                d.destination_hierarchy = mh_dst_hierarchy
                d.mapping = create_dictionary()

            # Specific code
            if mh_src_code in d.mapping:
                to_dict = d.mapping[mh_src_code]
            else:
                to_dict = create_dictionary()
            if mh_dst_code in to_dict:
                issues.append(Issue(itype=IType.ERROR,
                                    description="The mapping of '" + mh_src_code + "' into '" + mh_dst_code + "' has been already defined",
                                    location=IssueLocation(sheet_name=name, row=r, column=None)))
                return
            else:
                to_dict[mh_dst_code] = (mh_weight, r)  # NOTE: This could be an object instead of just a FLOAT or expression
                d.mapping[mh_src_code] = to_dict

        issues = []
        glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
        name = self._content["command_name"]

        local_mappings = create_dictionary()

        # Process parsed information
        for line in self._content["items"]:
            r = line["_row"]
            # If the line contains a reference to a dataset or hierarchy, expand it
            # If not, process it directly
            is_expansion = False
            if is_expansion:
                # TODO Iterate through dataset and/or hierarchy elements, producing a list of new items
                pass
            else:
                process_line(line)

        # Mappings post-processing
        for d in local_mappings:
            # Convert the mapping into:
            # [{"o": "", "to": [{"d": "", "w": ""}]}]
            # [ {o: origin category, to: [{d: destination category, w: weight assigned to destination category}] } ]
            mapping = []
            ds_rows = []  # Rows in which a dataset is mentioned
            for orig in local_mappings[d].mapping:
                lst = []
                for dst in local_mappings[d].mapping[orig]:
                    t = local_mappings[d].mapping[orig][dst]
                    lst.append(dict(d=dst, w=t[0]))
                    if local_mappings[d].origin_dataset:
                        ds_rows.append(t[1])
                mapping.append(dict(o=orig, to=lst))
            from nexinfosys.ie_imports.data_source_manager import DataSourceManager
            if local_mappings[d].origin_dataset:
                if not DataSourceManager.obtain_dataset_source(local_mappings[d].origin_dataset, datasets):
                    for r in ds_rows:
                        issues.append(Issue(itype=IType.ERROR,
                                            description=f"The dataset '{local_mappings[d].origin_dataset}' was not found",
                                            location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue
                dims, attrs, meas = obtain_dataset_metadata(local_mappings[d].origin_dataset, None, datasets)
                if local_mappings[d].origin_hierarchy not in dims:
                    issues.append(Issue(itype=IType.ERROR,
                                        description="The origin dimension '" + local_mappings[d].origin_hierarchy + "' does not exist in dataset '" + local_mappings[d].origin_dataset + "'",
                                        location=IssueLocation(sheet_name=name, row=r, column=None)))
                    continue
                else:
                    dim = dims[local_mappings[d].origin_hierarchy]
                    mapping = fill_map_with_all_origin_categories(dim, mapping)
            #
            origin_dataset = local_mappings[d].origin_dataset
            origin_hierarchy = local_mappings[d].origin_hierarchy
            destination_hierarchy = local_mappings[d].destination_hierarchy
            # Create Mapping and add it to Case Study mappings variable
            mappings[d] = Mapping(d, DataSourceManager.obtain_dataset_source(origin_dataset, datasets), origin_dataset, origin_hierarchy, destination_hierarchy, mapping)

        # TODO
        # Use the function to perform many to many mappings, "augment_dataframe_with_mapped_columns"
        # Put it to work !!!

        # One or more mapping in sequence could be specified?. The key is "source hierarchy+dest hierarchy"
        # Read mapping parameters

        return issues, None

    def estimate_execution_time(self):
        return 0

    def json_serialize(self):
        # Directly return the metadata dictionary
        return self._content

    def json_deserialize(self, json_input):
        # TODO Check validity
        issues = []
        if isinstance(json_input, dict):
            self._content = json_input
        else:
            self._content = json.loads(json_input)

        if "description" in json_input:
            self._description = json_input["description"]
        return issues



"""
Dataset with format:
variable original name
processor (just as a note)
interface (just as a note)
variable transformed name
value or expression
list of variables in the expression
scenario
time


For an scenario and time
* dictionary of changed name <-> original name
* Look for directly known variables
  - Add to dictionary (transform name to that in the expression)
* Parse expressions, look for variable names
* Transform expressions
* Prepare set of expressions
* Look for variables with known value
* Substitute
"""
from nexinfosys.model_services import State, get_case_study_registry_objects


def preprocess_expressions(state: State):
    """

    :param state:
    :return:
    """

    glb_idx, p_sets, hh, datasets, mappings = get_case_study_registry_objects(state)
    #